"""Utility functions to copy raw data files.

When writing BIDS datasets, we often move and/or rename raw data files. several
original data formats have properties that restrict such operations. That is,
moving/renaming raw data files naively might lead to broken files, for example
due to internal pointers that are not being updated.

"""

# Authors: Mainak Jas <mainak.jas@telecom-paristech.fr>
#          Alexandre Gramfort <alexandre.gramfort@telecom-paristech.fr>
#          Teon Brooks <teon.brooks@gmail.com>
#          Chris Holdgraf <choldgraf@berkeley.edu>
#          Stefan Appelhoff <stefan.appelhoff@mailbox.org>
#          Matt Sanderson <matt.sanderson@mq.edu.au>
#
# License: BSD-3-Clause
import os
import os.path as op
import re
import shutil as sh
from pathlib import Path

import numpy as np
from mne.io import anonymize_info, read_raw_bdf, read_raw_brainvision, read_raw_edf
from mne.utils import logger, verbose
from scipy.io import loadmat, savemat

from mne_bids.path import BIDSPath, _mkdir_p, _parse_ext
from mne_bids.utils import _check_anonymize, _get_mrk_meas_date, warn


def _copytree(src, dst, **kwargs):
    """See: https://github.com/jupyterlab/jupyterlab/pull/5150."""
    try:
        sh.copytree(src, dst, **kwargs)
    except sh.Error as error:
        # `copytree` throws an error if copying to + from NFS even though
        # the copy is successful (see https://bugs.python.org/issue24564)
        if "[Errno 22]" not in str(error) or not op.exists(dst):
            raise


def _get_brainvision_encoding(vhdr_file):
    """Get the encoding of .vhdr and .vmrk files.

    Parameters
    ----------
    vhdr_file : str
        Path to the header file.

    Returns
    -------
    enc : str
        Encoding of the .vhdr file to pass it on to open() function
        either 'UTF-8' (default) or whatever encoding scheme is specified
        in the header.

    """
    with open(vhdr_file, "rb") as ef:
        enc = ef.read()
        if enc.find(b"Codepage=") != -1:
            enc = enc[enc.find(b"Codepage=") + 9 :]
            enc = enc.split()[0]
            enc = enc.decode()
            src = "(read from header)"
        else:
            enc = "UTF-8"
            src = "(default)"
        logger.debug(f"Detected file encoding: {enc} {src}.")
    return enc


def _get_brainvision_paths(vhdr_path):
    """Get the .eeg/.dat and .vmrk file paths from a BrainVision header file.

    Parameters
    ----------
    vhdr_path : str
        Path to the header file.

    Returns
    -------
    paths : tuple
        Paths to the .eeg/.dat file at index 0 and the .vmrk file at index 1 of
        the returned tuple.

    """
    fname, ext = _parse_ext(vhdr_path)
    if ext != ".vhdr":
        raise ValueError(f'Expecting file ending in ".vhdr",' f" but got {ext}")

    # Header file seems fine
    # extract encoding from brainvision header file, or default to utf-8
    enc = _get_brainvision_encoding(vhdr_path)

    # ..and read it
    with open(vhdr_path, encoding=enc) as f:
        lines = f.readlines()

    # Try to find data file .eeg/.dat
    eeg_file_match = re.search(r"DataFile=(.*\.(eeg|dat))", " ".join(lines))

    if not eeg_file_match:
        raise ValueError("Could not find a .eeg or .dat file link in" f" {vhdr_path}")
    else:
        eeg_file = eeg_file_match.groups()[0]

    # Try to find marker file .vmrk
    vmrk_file_match = re.search(r"MarkerFile=(.*\.vmrk)", " ".join(lines))
    if not vmrk_file_match:
        raise ValueError("Could not find a .vmrk file link in" f" {vhdr_path}")
    else:
        vmrk_file = vmrk_file_match.groups()[0]

    # Make sure we are dealing with file names as is customary, not paths
    # Paths are problematic when copying the files to another system. Instead,
    # always use the file name and keep the file triplet in the same directory
    assert os.sep not in eeg_file
    assert os.sep not in vmrk_file

    # Assert the paths exist
    head, tail = op.split(vhdr_path)
    eeg_file_path = op.join(head, eeg_file)
    vmrk_file_path = op.join(head, vmrk_file)
    assert op.exists(eeg_file_path)
    assert op.exists(vmrk_file_path)

    # Return the paths
    return (eeg_file_path, vmrk_file_path)


def copyfile_ctf(src, dest):
    """Copy and rename CTF files to a new location.

    Parameters
    ----------
    src : path-like
        Path to the source raw .ds folder.
    dest : path-like
        Path to the destination of the new bids folder.

    See Also
    --------
    copyfile_brainvision
    copyfile_bti
    copyfile_edf
    copyfile_eeglab
    copyfile_kit

    """
    _copytree(src, dest)
    # list of file types to rename
    file_types = (
        ".acq",
        ".eeg",
        ".dat",
        ".hc",
        ".hist",
        ".infods",
        ".bak",
        ".meg4",
        ".newds",
        ".res4",
    )

    # Consider CTF files that are split having consecutively numbered extensions
    extra_ctf_file_types = tuple(
        f".{i}_meg4" for i in range(1, 21)
    )  # cap at 20 is arbitrary
    file_types += extra_ctf_file_types

    # Rename files in dest with the name of the dest directory
    fnames = [f for f in os.listdir(dest) if f.endswith(file_types)]
    bids_folder_name = op.splitext(op.split(dest)[-1])[0]
    for fname in fnames:
        ext = op.splitext(fname)[-1]
        os.replace(op.join(dest, fname), op.join(dest, bids_folder_name + ext))


def copyfile_kit(src, dest, subject_id, session_id, task, run, _init_kwargs):
    """Copy and rename KIT files to a new location.

    Parameters
    ----------
    src : path-like
        Path to the source raw .con or .sqd folder.
    dest : path-like
        Path to the destination of the new bids folder.
    subject_id : str | None
        The subject ID. Corresponds to "sub".
    session_id : str | None
        The session identifier. Corresponds to "ses".
    task : str | None
        The task identifier. Corresponds to "task".
    run : int | None
        The run number. Corresponds to "run".
    _init_kwargs : dict
        Extract information of marker and headpoints

    See Also
    --------
    copyfile_brainvision
    copyfile_bti
    copyfile_ctf
    copyfile_edf
    copyfile_eeglab

    """
    # create parent directories in case it does not exist yet
    _mkdir_p(op.dirname(dest))

    # KIT data requires the marker file to be copied over too
    sh.copyfile(src, dest)
    data_path = op.split(dest)[0]
    datatype = "meg"

    if "mrk" in _init_kwargs and _init_kwargs["mrk"] is not None:
        hpi = _init_kwargs["mrk"]
        acq_map = dict()
        if isinstance(hpi, list):
            if _get_mrk_meas_date(hpi[0]) > _get_mrk_meas_date(hpi[1]):
                raise ValueError("Markers provided in incorrect order.")
            _, marker_ext = _parse_ext(hpi[0])
            acq_map = dict(zip(["pre", "post"], hpi))
        else:
            _, marker_ext = _parse_ext(hpi)
            acq_map[None] = hpi
        for key, value in acq_map.items():
            marker_path = BIDSPath(
                subject=subject_id,
                session=session_id,
                task=task,
                run=run,
                acquisition=key,
                suffix="markers",
                extension=marker_ext,
                datatype=datatype,
            )
            sh.copyfile(value, op.join(data_path, marker_path.basename))

    for acq in ["elp", "hsp"]:
        if acq in _init_kwargs and _init_kwargs[acq] is not None:
            position_file = _init_kwargs[acq]
            task, run, acq = None, None, acq.upper()
            position_ext = ".pos"
            position_path = BIDSPath(
                subject=subject_id,
                session=session_id,
                task=task,
                run=run,
                acquisition=acq,
                suffix="headshape",
                extension=position_ext,
                datatype=datatype,
            )
            sh.copyfile(position_file, op.join(data_path, position_path.basename))


def _replace_file(fname, pattern, replace):
    """Overwrite file, replacing end of lines matching pattern with replace."""
    new_content = []
    for line in open(fname):
        match = re.match(pattern, line)
        if match:
            line = match.group()[: -len(replace)] + replace + "\n"
        new_content.append(line)

    with open(fname, "w", encoding="utf-8") as fout:
        fout.writelines(new_content)


def _anonymize_brainvision(vhdr_file, date):
    """Anonymize vmrk and vhdr files in place using `date` datetime object."""
    _, vmrk_file = _get_brainvision_paths(vhdr_file)

    # Go through VMRK
    pattern = re.compile(r"^Mk\d+=New Segment,.*,\d+,\d+,\d+,\d{20}$")
    replace = date.strftime("%Y%m%d%H%M%S%f")
    _replace_file(vmrk_file, pattern, replace)

    # Go through VHDR
    pattern = re.compile(r"^Impedance \[kOhm\] at \d\d:\d\d:\d\d :$")
    replace = f'at {date.strftime("%H:%M:%S")} :'
    _replace_file(vhdr_file, pattern, replace)


@verbose
def copyfile_brainvision(vhdr_src, vhdr_dest, anonymize=None, verbose=None):
    """Copy a BrainVision file triplet to a new location and repair links.

    The BrainVision file format consists of three files:
    .vhdr, .eeg/.dat, and .vmrk
    The .eeg/.dat and .vmrk files associated with the .vhdr file will be
    given names as in `vhdr_dest` with adjusted extensions. Internal file
    pointers will be fixed.

    Parameters
    ----------
    vhdr_src : path-like
        The source path of the .vhdr file to be copied.
    vhdr_dest : path-like
        The destination path of the .vhdr file.
    anonymize : dict | None
        If None (default), no anonymization is performed.
        If dict, data will be anonymized depending on the keys provided with
        the dict: `daysback` is a required key, `keep_his` is an optional key.

        `daysback` : int
            Number of days by which to move back the recording date in time.
            In studies with multiple subjects the relative recording date
            differences between subjects can be kept by using the same number
            of `daysback` for all subject anonymizations. `daysback` should be
            great enough to shift the date prior to 1925 to conform with BIDS
            anonymization rules.

        `keep_his` : bool
            By default (False), all subject information next to the recording
            date will be overwritten as well. If True, keep subject information
            apart from the recording date.

    %(verbose)s

    See Also
    --------
    mne.io.anonymize_info
    copyfile_bti
    copyfile_ctf
    copyfile_edf
    copyfile_eeglab
    copyfile_kit

    """
    # Get extension of the brainvision file
    fname_src, ext_src = _parse_ext(vhdr_src)
    fname_dest, ext_dest = _parse_ext(vhdr_dest)
    if ext_src != ext_dest:
        raise ValueError(
            f"Need to move data with same extension, "
            f' but got "{ext_src}" and "{ext_dest}"'
        )

    eeg_file_path, vmrk_file_path = _get_brainvision_paths(vhdr_src)

    # extract encoding from brainvision header file, or default to utf-8
    enc = _get_brainvision_encoding(vhdr_src)

    # raise warning if binary file has .dat extension
    if ".dat" in eeg_file_path:
        warn(
            "The file extension of your binary EEG data file is .dat, while "
            "the expected extension for raw data is .eeg. "
            "This might imply it's preprocessed or processed data: "
            "We copied the files and changed the extension to .eeg, "
            "but please ensure that this is actually BIDS compatible data!"
        )

    # Copy data .eeg/.dat ... no links to repair
    sh.copyfile(eeg_file_path, fname_dest + ".eeg")

    # Write new header and marker files, fixing the file pointer links
    # For that, we need to replace an old "basename" with a new one
    # assuming that all .eeg/.dat, .vhdr, .vmrk share one basename
    __, basename_src = op.split(fname_src)
    assert op.split(eeg_file_path)[-1] in [basename_src + ".eeg", basename_src + ".dat"]
    assert basename_src + ".vmrk" == op.split(vmrk_file_path)[-1]
    __, basename_dest = op.split(fname_dest)
    search_lines = [
        "DataFile=" + basename_src + ".eeg",
        "DataFile=" + basename_src + ".dat",
        "MarkerFile=" + basename_src + ".vmrk",
    ]

    with open(vhdr_src, encoding=enc) as fin:
        with open(vhdr_dest, "w", encoding=enc) as fout:
            for line in fin.readlines():
                if line.strip() in search_lines:
                    line = line.replace(basename_src, basename_dest)
                fout.write(line)

    with open(vmrk_file_path, encoding=enc) as fin:
        with open(fname_dest + ".vmrk", "w", encoding=enc) as fout:
            for line in fin.readlines():
                if line.strip() in search_lines:
                    line = line.replace(basename_src, basename_dest)
                fout.write(line)

    if anonymize is not None:
        raw = read_raw_brainvision(vhdr_src, preload=False, verbose=0)
        daysback, keep_his, _ = _check_anonymize(anonymize, raw, ".vhdr")
        raw.info = anonymize_info(raw.info, daysback=daysback, keep_his=keep_his)
        _anonymize_brainvision(fname_dest + ".vhdr", date=raw.info["meas_date"])

    for ext in [".eeg", ".vhdr", ".vmrk"]:
        _, fname = os.path.split(fname_dest + ext)
        dirname = op.dirname(op.realpath(vhdr_dest))
        logger.info(f'Created "{fname}" in "{dirname}".')
    if anonymize:
        logger.info("Anonymized all dates in VHDR and VMRK.")


def copyfile_edf(src, dest, anonymize=None):
    """Copy an EDF, EDF+, or BDF file to a new location, optionally anonymize.

    .. warning:: EDF/EDF+/BDF files contain two fields for recording dates:
                 A generic "startdate" field that supports only 2-digit years,
                 and a "Startdate" field as part of the "local recording
                 identification", which supports 4-digit years.
                 If you want to anonymize your file, MNE-BIDS will set the
                 "startdate" field to 85 (i.e., 1985), the earliest possible
                 date for that field. However, the "Startdate" field in the
                 file's "local recording identification" and the date in the
                 session's corresponding ``scans.tsv`` will be set correctly
                 according to the argument provided to the ``anonymize``
                 parameter. Note that it is possible that not all EDF/EDF+/BDF
                 reading software parses the accurate recording date, and
                 that for some reading software, the wrong year (1985) may
                 be parsed.

    Parameters
    ----------
    src : path-like
        The source path of the .edf or .bdf file to be copied.
    dest : path-like
        The destination path of the .edf or .bdf file.
    anonymize : dict | None
        If None (default), no anonymization is performed.
        If dict, data will be anonymized depending on the keys provided with
        the dict: `daysback` is a required key, `keep_his` is an optional key.

        `daysback` : int
            Number of days by which to move back the recording date in time.
            In studies with multiple subjects the relative recording date
            differences between subjects can be kept by using the same number
            of `daysback` for all subject anonymizations. `daysback` should be
            great enough to shift the date prior to 1925 to conform with BIDS
            anonymization rules. Due to limitations of the EDF/BDF format, the
            year of the anonymized date will always be set to 1985 in the
            'startdate' field of the file. The correctly-shifted year will be
            written to the 'local recording identification' region of the
            file header, which may not be parsed by all EDF/EDF+/BDF reader
            software.

        `keep_his` : bool
            By default (False), all subject information next to the recording
            date will be overwritten as well. If True, keep subject information
            apart from the recording date. Participant names and birthdates
            will always be anonymized if present, regardless of this setting.

    See Also
    --------
    mne.io.anonymize_info
    copyfile_brainvision
    copyfile_bti
    copyfile_ctf
    copyfile_eeglab
    copyfile_kit

    """
    # Ensure source & destination extensions are the same
    fname_src, ext_src = _parse_ext(src)
    fname_dest, ext_dest = _parse_ext(dest)

    if ext_src.lower() != ext_dest.lower():
        raise ValueError(
            f"Need to move data with same extension, "
            f' but got "{ext_src}" and "{ext_dest}"'
        )

    if ext_dest in [".EDF", ".BDF"]:
        warn(
            "Upper-case extension for EDF/BDF files is not supported "
            "in BIDS. Converting destination extension to lower-case."
        )
        ext_dest = ext_dest.lower()
        dest = Path(dest).with_suffix(ext_dest)

    # Copy data prior to any anonymization
    sh.copyfile(src, dest)

    # Anonymize EDF/BDF data, if requested
    if anonymize is not None:
        if ext_src in [".bdf", ".BDF"]:
            raw = read_raw_bdf(dest, preload=False, verbose=0)
        elif ext_src in [".edf", ".EDF"]:
            raw = read_raw_edf(dest, preload=False, verbose=0)
        else:
            raise ValueError(f"Unsupported file type ({ext_src})")

        # Get subject info, recording info, and recording date
        with open(dest, "rb") as f:
            f.seek(8)  # id_info field starts 8 bytes in
            id_info = f.read(80).decode("ascii").rstrip()
            rec_info = f.read(80).decode("ascii").rstrip()

        # Parse metadata from file
        if len(id_info) == 0 or len(id_info.split(" ")) != 4:
            id_info = "X X X X"
        if len(rec_info) == 0 or len(rec_info.split(" ")) != 5:
            rec_info = "Startdate X X X X"
        pid, sex, birthdate, name = id_info.split(" ")
        start_date, admin_code, tech, equip = rec_info.split(" ")[1:5]

        # Try to anonymize the recording date
        daysback, keep_his, _ = _check_anonymize(anonymize, raw, ".edf")
        anonymize_info(raw.info, daysback=daysback, keep_his=keep_his)
        start_date = "01-JAN-1985"
        meas_date = "01.01.85"

        # Anonymize ID info and write to file
        if keep_his:
            # Always remove participant birthdate and name to be safe
            id_info = [pid, sex, "X", "X"]
            rec_info = ["Startdate", start_date, admin_code, tech, equip]
        else:
            id_info = ["0", "X", "X", "X"]
            rec_info = ["Startdate", start_date, "X", "mne-bids_anonymize", "X"]
        with open(dest, "r+b") as f:
            f.seek(8)  # id_info field starts 8 bytes in
            f.write(bytes(" ".join(id_info).ljust(80), "ascii"))
            f.write(bytes(" ".join(rec_info).ljust(80), "ascii"))
            f.write(bytes(meas_date, "ascii"))


def copyfile_eeglab(src, dest):
    """Copy an EEGLAB file to a new location.

    If the EEGLAB ``.set`` file comes with an accompanying ``.fdt`` binary file
    that contains the actual data, this function will copy this file, too, and
    update all internal pointers in the new ``.set`` file.

    Parameters
    ----------
    src : path-like
        Path to the source raw .set file.
    dest : path-like
        Path to the destination of the new .set file.

    See Also
    --------
    copyfile_brainvision
    copyfile_bti
    copyfile_ctf
    copyfile_edf
    copyfile_kit

    """
    # Get extension of the EEGLAB file
    _, ext_src = _parse_ext(src)
    fname_dest, ext_dest = _parse_ext(dest)
    if ext_src != ext_dest:
        raise ValueError(
            f"Need to move data with same extension" f" but got {ext_src}, {ext_dest}"
        )

    # Load the EEG struct
    # NOTE: do *not* simplify cells, because this changes the underlying
    # structure and potentially breaks re-reading of the file
    uint16_codec = None
    eeg = loadmat(
        file_name=src,
        simplify_cells=False,
        appendmat=False,
        uint16_codec=uint16_codec,
        mat_dtype=True,
    )
    oldstyle = False
    if "EEG" in eeg:
        eeg = eeg["EEG"]
        oldstyle = True

    has_fdt_link = False
    try:
        # If the data field is a string, it points to a .fdt file in src dir
        if isinstance(eeg["data"][0, 0][0], str):
            has_fdt_link = True
    except IndexError:
        pass

    if has_fdt_link:
        fdt_fname = eeg["data"][0, 0][0]

        assert fdt_fname.endswith(".fdt"), f"Unexpected fdt name: {fdt_fname}"
        head, _ = op.split(src)
        fdt_path = op.join(head, fdt_fname)

        # Copy the .fdt file and give it a new name
        fdt_name_new = fname_dest + ".fdt"
        sh.copyfile(fdt_path, fdt_name_new)

        # Now adjust the pointer in the .set file
        # NOTE: Clunky numpy code is to match MATLAB structure for "savemat"
        _, tail = op.split(fdt_name_new)
        new_value = np.empty((1, 1), dtype=object)
        new_value[0, 0] = np.atleast_1d(np.array(tail))
        eeg["data"] = new_value

        # Save the EEG dictionary as a Matlab struct again
        mdict = dict(EEG=eeg) if oldstyle else eeg
        savemat(file_name=dest, mdict=mdict, appendmat=False)
    else:
        # If no .fdt file, simply copy the .set file, no modifications
        # necessary
        sh.copyfile(src, dest)


def copyfile_bti(raw, dest):
    """Copy BTi data.

    Parameters
    ----------
    raw : mne.io.Raw
        An MNE-Python raw object of BTi data.
    dest : path-like
        Destination to copy the BTi data to.

    See Also
    --------
    copyfile_brainvision
    copyfile_ctf
    copyfile_edf
    copyfile_eeglab
    copyfile_kit

    """
    os.makedirs(dest, exist_ok=True)
    for key, val in (
        ("pdf_fname", None),
        ("config_fname", "config"),
        ("head_shape_fname", "hs_file"),
    ):
        keyfile = raw._raw_extras[0].get(key)

        # Keep name of pdf file
        if key == "pdf_fname":
            val = op.basename(keyfile)

        # If no headshape file present, cannot copy it
        if key == "head_shape_fname" and keyfile is None:
            continue

        sh.copyfile(keyfile, op.join(dest, val))



r"""Inspect MEG and EEG raw data, and interactively mark channels as bad.

example usage:
$ mne_bids inspect --subject_id=01 --task=experiment --session=test \
--datatype=meg --suffix=meg --bids_root=bids_root

"""
# Authors: Richard Höchenberger <richard.hoechenberger@gmail.com>
#
# License: BSD-3-Clause

from mne.utils import logger

import mne_bids
from mne_bids import BIDSPath, inspect_dataset


def run():
    """Run the mark_channels command."""
    from mne.commands.utils import get_optparser

    parser = get_optparser(
        __file__,
        usage="usage: %prog options args",
        prog_prefix="mne_bids",
        version=mne_bids.__version__,
    )

    parser.add_option(
        "--bids_root",
        dest="bids_root",
        help="The path of the folder containing the BIDS dataset",
    )
    parser.add_option("--subject_id", dest="subject", help="Subject name")
    parser.add_option("--session_id", dest="session", help="Session name")
    parser.add_option("--task", dest="task", help="Task name")
    parser.add_option("--acq", dest="acquisition", help="Acquisition parameter")
    parser.add_option("--run", dest="run", help="Run number")
    parser.add_option("--proc", dest="processing", help="Processing label.")
    parser.add_option("--rec", dest="recording", help="Recording name")
    parser.add_option(
        "--type", dest="datatype", help="Recording data type, e.g. meg, ieeg or eeg"
    )
    parser.add_option(
        "--suffix",
        dest="suffix",
        help="The filename suffix, i.e. the last part before the extension",
    )
    parser.add_option(
        "--ext",
        dest="extension",
        help="The filename extension, including the leading period, e.g. .fif",
    )
    parser.add_option(
        "--find_flat",
        dest="find_flat",
        help="Whether to auto-detect flat channels and time segments",
    )
    parser.add_option(
        "--l_freq", dest="l_freq", help="The high-pass filter cutoff frequency"
    )
    parser.add_option(
        "--h_freq", dest="h_freq", help="The low-pass filter cutoff frequency"
    )
    parser.add_option(
        "--verbose",
        dest="verbose",
        action="store_true",
        help="Whether do generate additional diagnostic output",
    )

    opt, args = parser.parse_args()
    if args:
        parser.print_help()
        parser.error(
            f"Please do not specify arguments without flags. " f"Got: {args}.\n"
        )

    if opt.bids_root is None:
        parser.print_help()
        parser.error("You must specify bids_root")

    bids_path = BIDSPath(
        subject=opt.subject,
        session=opt.session,
        task=opt.task,
        acquisition=opt.acquisition,
        run=opt.run,
        processing=opt.processing,
        recording=opt.recording,
        datatype=opt.datatype,
        suffix=opt.suffix,
        extension=opt.extension,
        root=opt.bids_root,
    )

    find_flat = True if opt.find_flat is None else bool(opt.find_flat)
    l_freq = None if opt.l_freq is None else float(opt.l_freq)
    h_freq = None if opt.h_freq is None else float(opt.h_freq)

    logger.info(f"Inspecting {bids_path.basename} …")
    inspect_dataset(
        bids_path=bids_path,
        find_flat=find_flat,
        l_freq=l_freq,
        h_freq=h_freq,
        verbose=opt.verbose,
    )


if __name__ == "__main__":
    run()



"""Rename files (making a copy) and update their internal pointers.

example usage: $ mne_bids cp --input myfile.vhdr --output sub-01_task-test.vhdr
"""

# Authors: Stefan Appelhoff <stefan.appelhoff@mailbox.org>
#
# License: BSD-3-Clause
import mne_bids
from mne_bids.copyfiles import copyfile_brainvision, copyfile_ctf, copyfile_eeglab


def run():
    """Run the cp command."""
    from mne.commands.utils import get_optparser

    accepted_formats_msg = "(accepted formats: BrainVision .vhdr, EEGLAB .set, CTF .ds)"

    parser = get_optparser(
        __file__,
        usage="usage: %prog -i INPUT -o OUTPUT",
        prog_prefix="mne_bids",
        version=mne_bids.__version__,
    )

    parser.add_option(
        "-i",
        "--input",
        dest="input",
        help=f"path to the input file. {accepted_formats_msg}",
        metavar="INPUT",
    )

    parser.add_option(
        "-o",
        "--output",
        dest="output",
        help="path to the output file (MUST be same format as input file)",
        metavar="OUTPUT",
    )

    parser.add_option(
        "-v",
        "--verbose",
        dest="verbose",
        help="set logging level to verbose",
        action="store_true",
    )

    opt, args = parser.parse_args()
    opt_dict = vars(opt)

    # Check the usage and raise error if invalid
    if len(args) > 0:
        parser.print_help()
        parser.error(
            f'Do not specify arguments without flags. Found: "{args}".\n'
            "Did you forget to provide -i and -o?"
        )

    if not opt_dict.get("input") or not opt_dict.get("output"):
        parser.print_help()
        parser.error(
            "Incorrect number of arguments. Supply one input and one "
            f'output file. You supplied: "{opt}"'
        )

    # Attempt to do the copying. Errors will be raised by the copyfile
    # functions if there are issues with the file formats
    if opt.input.endswith(".vhdr"):
        copyfile_brainvision(opt.input, opt.output, opt.verbose)
    elif opt.input.endswith(".set"):
        copyfile_eeglab(opt.input, opt.output)
    elif opt.input.endswith(".ds"):
        copyfile_ctf(opt.input, opt.output)
    else:
        parser.error(f'{accepted_formats_msg} You supplied: "{opt}"')


if __name__ == "__main__":
    run()



"""Utility and helper functions for MNE-BIDS."""

# Authors: Mainak Jas <mainak.jas@telecom-paristech.fr>
#          Alexandre Gramfort <alexandre.gramfort@telecom-paristech.fr>
#          Teon Brooks <teon.brooks@gmail.com>
#          Chris Holdgraf <choldgraf@berkeley.edu>
#          Stefan Appelhoff <stefan.appelhoff@mailbox.org>
#          Matt Sanderson <matt.sanderson@mq.edu.au>
#
# License: BSD-3-Clause
import json
import os
import re
from datetime import date, datetime, timedelta, timezone
from os import path as op

import numpy as np
from mne import pick_types
from mne.channels import make_standard_montage
from mne.io.kit.kit import get_kit_info
from mne.utils import logger, verbose
from mne.utils import warn as _warn

from mne_bids.tsv_handler import _to_tsv

# This regex matches key-val pairs. Any characters are allowed in the key and
# the value, except these special symbols: - _ . \ /
param_regex = re.compile(r"([^-_\.\\\/]+)-([^-_\.\\\/]+)")


def _ensure_tuple(x):
    """Return a tuple."""
    if x is None:
        return tuple()
    elif isinstance(x, str):
        return (x,)
    else:
        return tuple(x)


def _get_ch_type_mapping(fro="mne", to="bids"):
    """Map between BIDS and MNE nomenclatures for channel types.

    Parameters
    ----------
    fro : str
        Mapping from nomenclature of `fro`. Can be 'mne', 'bids'
    to : str
        Mapping to nomenclature of `to`. Can be 'mne', 'bids'

    Returns
    -------
    mapping : dict
        Dictionary mapping from one nomenclature of channel types to another.
        If a key is not present, a default value will be returned that depends
        on the `fro` and `to` parameters.

    Notes
    -----
    For the mapping from BIDS to MNE, MEG channel types are ignored for now.
    Furthermore, this is not a one-to-one mapping: Incomplete and partially
    one-to-many/many-to-one.

    Bio channels are supported in mne-python and are converted to MISC
    because there is no "Bio" supported channel in BIDS.
    """
    if fro == "mne" and to == "bids":
        mapping = dict(
            eeg="EEG",
            misc="MISC",
            stim="TRIG",
            emg="EMG",
            ecog="ECOG",
            seeg="SEEG",
            eog="EOG",
            ecg="ECG",
            resp="RESP",
            bio="MISC",
            dbs="DBS",
            gsr="GSR",
            temperature="TEMP",
            # NIRS
            fnirs_cw_amplitude="NIRSCWAMPLITUDE",
            # MEG channels
            meggradaxial="MEGGRADAXIAL",
            megmag="MEGMAG",
            megrefgradaxial="MEGREFGRADAXIAL",
            meggradplanar="MEGGRADPLANAR",
            megrefmag="MEGREFMAG",
            ias="MEGOTHER",
            syst="MEGOTHER",
            exci="MEGOTHER",
        )

    elif fro == "bids" and to == "mne":
        mapping = dict(
            EEG="eeg",
            MISC="misc",
            TRIG="stim",
            EMG="emg",
            ECOG="ecog",
            SEEG="seeg",
            EOG="eog",
            ECG="ecg",
            RESP="resp",
            GSR="gsr",
            TEMP="temperature",
            # NIRS
            NIRSCWAMPLITUDE="fnirs_cw_amplitude",
            NIRS="fnirs_cw_amplitude",
            # No MEG channels for now (see Notes above)
            # Many to one mapping
            VEOG="eog",
            HEOG="eog",
            DBS="dbs",
        )
    else:
        raise ValueError(
            "Only two types of mappings are currently supported: "
            "from mne to bids, or from bids to mne. However, "
            f'you specified from "{fro}" to "{to}"'
        )

    return mapping


def _handle_datatype(raw, datatype):
    """Check if datatype exists in raw object or infer datatype if possible.

    Parameters
    ----------
    raw : mne.io.Raw
        Raw object.
    datatype : str | None
        Can be one of either ``'meg'``, ``'eeg'``, or ``'ieeg'``. If ``None``,
        `mne.utils._handle_datatype()` will attempt to infer the datatype from
        the ``raw`` object. In case of multiple data types in the ``raw``
        object, ``datatype`` must not be ``None``.

    Returns
    -------
    datatype : str
        One of either ``'meg'``, ``'eeg'``, or ``'ieeg'``.
    """
    if datatype is not None:
        _check_datatype(raw, datatype)
        # MEG data is not supported by BrainVision or EDF files
        if datatype in ["eeg", "ieeg"] and "meg" in raw:
            logger.info(
                f"{os.linesep}Both {datatype} and 'meg' data found. "
                f"BrainVision and EDF do not support 'meg' data. "
                f"The data will therefore be stored as 'meg' data. "
                f"If you wish to store your {datatype} data in "
                f"BrainVision or EDF, please remove the 'meg'"
                f"channels from your recording.{os.linesep}"
            )
            datatype = "meg"
    else:
        datatypes = list()
        ieeg_types = ["seeg", "ecog", "dbs"]
        if any(ieeg_type in raw for ieeg_type in ieeg_types):
            datatypes.append("ieeg")
        if "meg" in raw:
            datatypes.append("meg")
        if "eeg" in raw:
            datatypes.append("eeg")
        if "fnirs_cw_amplitude" in raw:
            datatypes.append("nirs")
        if len(datatypes) == 0:
            raise ValueError(
                "No MEG, EEG or iEEG channels found in data. "
                "Please use raw.set_channel_types to set the "
                "channel types in the data."
            )
        elif len(datatypes) > 1:
            if "meg" in datatypes and "ieeg" not in datatypes:
                datatype = "meg"
            elif "ieeg" in datatypes and "meg" not in datatypes:
                datatype = "ieeg"
            else:
                raise ValueError(
                    f"Multiple data types (``{datatypes}``) were "
                    "found in the data. Please specify the "
                    "datatype using "
                    '`bids_path.update(datatype="<datatype>")` '
                    "or use raw.set_channel_types to set the "
                    "correct channel types in the raw object."
                )
        else:
            datatype = datatypes[0]
    return datatype


def _age_on_date(bday, exp_date):
    """Calculate age from birthday and experiment date.

    Parameters
    ----------
    bday : datetime.datetime
        The birthday of the participant.
    exp_date : datetime.datetime
        The date the experiment was performed on.

    """
    if exp_date < bday:
        raise ValueError("The experimentation date must be after the birth date")
    if exp_date.month > bday.month:
        return exp_date.year - bday.year
    elif exp_date.month == bday.month:
        if exp_date.day >= bday.day:
            return exp_date.year - bday.year
    return exp_date.year - bday.year - 1


def _check_types(variables):
    """Make sure all vars are str or None."""
    for var in variables:
        if not isinstance(var, (str, type(None))):
            raise ValueError(
                f"You supplied a value ({var}) of type "
                f"{type(var)}, where a string or None was "
                f"expected."
            )


def _write_json(fname, dictionary, overwrite=False):
    """Write JSON to a file."""
    if op.exists(fname) and not overwrite:
        raise FileExistsError(
            f'"{fname}" already exists. ' "Please set overwrite to True."
        )

    json_output = json.dumps(dictionary, indent=4)
    with open(fname, "w", encoding="utf-8") as fid:
        fid.write(json_output)
        fid.write("\n")

    logger.info(f"Writing '{fname}'...")


@verbose
def _write_tsv(fname, dictionary, overwrite=False, verbose=None):
    """Write an ordered dictionary to a .tsv file."""
    if op.exists(fname) and not overwrite:
        raise FileExistsError(
            f'"{fname}" already exists. ' "Please set overwrite to True."
        )
    _to_tsv(dictionary, fname)

    logger.info(f"Writing '{fname}'...")


def _write_text(fname, text, overwrite=False):
    """Write text to a file."""
    if op.exists(fname) and not overwrite:
        raise FileExistsError(
            f'"{fname}" already exists. ' "Please set overwrite to True."
        )
    with open(fname, "w", encoding="utf-8-sig") as fid:
        fid.write(text)
        fid.write("\n")

    logger.info(f"Writing '{fname}'...")


def _check_key_val(key, val):
    """Perform checks on a value to make sure it adheres to the spec."""
    if any(ii in val for ii in ["-", "_", "/"]):
        raise ValueError(
            "Unallowed `-`, `_`, or `/` found in key/value pair" f" {key}: {val}"
        )
    return key, val


def _get_mrk_meas_date(mrk):
    """Find the measurement date from a KIT marker file."""
    info = get_kit_info(mrk, False)[0]
    meas_date = info.get("meas_date", None)
    if isinstance(meas_date, (tuple, list, np.ndarray)):
        meas_date = meas_date[0]
    if isinstance(meas_date, datetime):
        meas_datetime = meas_date
    elif meas_date is not None:
        meas_datetime = datetime.fromtimestamp(meas_date)
    else:
        meas_datetime = datetime.min
    return meas_datetime


def _infer_eeg_placement_scheme(raw):
    """Based on the channel names, try to infer an EEG placement scheme.

    Parameters
    ----------
    raw : mne.io.Raw
        The data as MNE-Python Raw object.

    Returns
    -------
    placement_scheme : str
        Description of the EEG placement scheme. Will be "n/a" for unsuccessful
        extraction.

    """
    placement_scheme = "n/a"
    # Check if the raw data contains eeg data at all
    if "eeg" not in raw:
        return placement_scheme

    # How many of the channels in raw are based on the extended 10/20 system
    sel = pick_types(raw.info, meg=False, eeg=True)
    ch_names = [raw.ch_names[i] for i in sel]
    channel_names = [ch.lower() for ch in ch_names]
    montage1005 = make_standard_montage("standard_1005")
    montage1005_names = [ch.lower() for ch in montage1005.ch_names]

    if set(channel_names).issubset(set(montage1005_names)):
        placement_scheme = "based on the extended 10/20 system"

    return placement_scheme


def _scale_coord_to_meters(coord, unit):
    """Scale units to meters (mne-python default)."""
    if unit == "cm":
        return np.divide(coord, 100.0)
    elif unit == "mm":
        return np.divide(coord, 1000.0)
    else:
        return coord


def _check_empty_room_basename(bids_path):
    if bids_path.subject != "emptyroom":
        return
    # only check task entity for emptyroom when it is the sidecar/MEG file
    if bids_path.suffix != "meg":
        return
    if bids_path.acquisition in ("calibration", "crosstalk"):
        return
    if bids_path.task != "noise":
        raise ValueError(
            f'task must be "noise" if subject is "emptyroom", but '
            f"received: {bids_path.task}"
        )


def _check_anonymize(anonymize, raw, ext):
    """Check the `anonymize` dict."""
    # if info['meas_date'] None, then the dates are not stored
    if raw.info["meas_date"] is None:
        daysback = None
    else:
        if "daysback" not in anonymize or anonymize["daysback"] is None:
            raise ValueError("`daysback` argument required to anonymize.")
        daysback = anonymize["daysback"]
        daysback_min, daysback_max = _get_anonymization_daysback(raw)
        if daysback < daysback_min:
            warn(
                "`daysback` is too small; the measurement date "
                "is after 1925, which is not recommended by BIDS."
                "The minimum `daysback` value for changing the "
                "measurement date of this data to before this date "
                f"is {daysback_min}"
            )
        if ext == ".fif" and daysback > daysback_max:
            raise ValueError(
                "`daysback` exceeds maximum value MNE "
                "is able to store in FIF format, must "
                f"be less than {daysback_max}"
            )
    keep_his = anonymize["keep_his"] if "keep_his" in anonymize else False
    keep_source = anonymize["keep_source"] if "keep_source" in anonymize else False
    return daysback, keep_his, keep_source


def _get_anonymization_daysback(raw):
    """Get the min and max number of daysback necessary to satisfy BIDS specs.

    Parameters
    ----------
    raw : mne.io.Raw
        Subject raw data.

    Returns
    -------
    daysback_min : int
        The minimum number of daysback necessary to be compatible with BIDS.
    daysback_max : int
        The maximum number of daysback that MNE can store.
    """
    this_date = _stamp_to_dt(raw.info["meas_date"]).date()
    daysback_min = (this_date - date(year=1924, month=12, day=31)).days
    daysback_max = (
        this_date
        - datetime.fromtimestamp(0).date()
        + timedelta(seconds=np.iinfo(">i4").max)
    ).days
    return daysback_min, daysback_max


@verbose
def get_anonymization_daysback(raws, verbose=None):
    """Get the group min and max number of daysback necessary for BIDS specs.

    .. warning:: It is important that you remember the anonymization
                 number if you would ever like to de-anonymize but
                 that it is not included in the code publication
                 as that would break the anonymization.

    BIDS requires that anonymized dates be before 1925. In order to
    preserve the longitudinal structure and ensure anonymization, the
    user is asked to provide the same `daysback` argument to each call
    of `write_raw_bids`. To determine the minimum number of daysback
    necessary, this function will calculate the minimum number based on
    the most recent measurement date of raw objects.

    Parameters
    ----------
    raw : mne.io.Raw | list of mne.io.Raw
        Subject raw data or list of raw data from several subjects.
    %(verbose)s

    Returns
    -------
    daysback_min : int
        The minimum number of daysback necessary to be compatible with BIDS.
    daysback_max : int
        The maximum number of daysback that MNE can store.
    """
    if not isinstance(raws, list):
        raws = list([raws])
    daysback_min_list = list()
    daysback_max_list = list()
    for raw in raws:
        if raw.info["meas_date"] is not None:
            daysback_min, daysback_max = _get_anonymization_daysback(raw)
            daysback_min_list.append(daysback_min)
            daysback_max_list.append(daysback_max)
    if not daysback_min_list or not daysback_max_list:
        raise ValueError(
            "All measurement dates are None, pass any `daysback` value to anonymize."
        )
    daysback_min = max(daysback_min_list)
    daysback_max = min(daysback_max_list)
    if daysback_min > daysback_max:
        raise ValueError(
            "The dataset spans more time than can be "
            "accomodated by MNE, you may have to "
            "not follow BIDS recommendations and use"
            "anonymized dates after 1925"
        )
    return daysback_min, daysback_max


def _stamp_to_dt(utc_stamp):
    """Convert POSIX timestamp to datetime object in Windows-friendly way."""
    # This is a windows datetime bug for timestamp < 0. A negative value
    # is needed for anonymization which requires the date to be moved back
    # to before 1925. This then requires a negative value of daysback
    # compared the 1970 reference date.
    if isinstance(utc_stamp, datetime):
        return utc_stamp
    stamp = [int(s) for s in utc_stamp]
    if len(stamp) == 1:  # In case there is no microseconds information
        stamp.append(0)
    return datetime.fromtimestamp(0, tz=timezone.utc) + timedelta(
        0, stamp[0], stamp[1]
    )  # day, sec, μs


def _check_datatype(raw, datatype):
    """Check if datatype exists in given raw object.

    Parameters
    ----------
    raw : mne.io.Raw
        Raw object.
    datatype : str
        Can be one of either ``'meg'``, ``'eeg'``, or ``'ieeg'``.

    Returns
    -------
    None
    """
    supported_types = ("meg", "eeg", "ieeg", "nirs")
    if datatype not in supported_types:
        raise ValueError(
            f"The specified datatype {datatype} is currently not supported. "
            f"It should be one of  either `meg`, `eeg` or `ieeg` (Got "
            f"`{datatype}`. Please specify a valid datatype using "
            f'`bids_path.update(datatype="<datatype>")`.'
        )
    datatype_matches = False
    if datatype == "eeg" and datatype in raw:
        datatype_matches = True
    elif datatype == "meg" and datatype in raw:
        datatype_matches = True
    elif datatype == "nirs" and "fnirs_cw_amplitude" in raw:
        datatype_matches = True
    elif datatype == "ieeg":
        ieeg_types = ("seeg", "ecog", "dbs")
        if any(ieeg_type in raw for ieeg_type in ieeg_types):
            datatype_matches = True
    if not datatype_matches:
        raise ValueError(
            f"The specified datatype {datatype} was not found in the raw "
            "object. Please specify the correct datatype using "
            '`bids_path.update(datatype="<datatype>")` or use '
            "raw.set_channel_types to set the correct channel types in "
            "the raw object."
        )


def _import_nibabel(why="work with MRI data"):
    try:
        import nibabel  # noqa
    except ImportError as exc:
        raise exc.__class__(
            f"nibabel is required to {why} but could not be imported, " f"got: {exc}"
        ) from None
    else:
        return nibabel


def warn(
    message,
    category=RuntimeWarning,
    module="mne_bids",
    ignore_namespaces=("mne", "mne_bids"),
):  # noqa: D103
    """Emit a warning."""
    _warn(
        message,
        category=category,
        module=module,
        ignore_namespaces=ignore_namespaces,
    )


# Some of the defaults here will be wrong but it should be close enough
warn.__doc__ = getattr(_warn, "__doc__", None)



"""Inspect and annotate BIDS raw data."""
# Authors: Richard Höchenberger <richard.hoechenberger@gmail.com>
#          Stefan Appelhoff <stefan.appelhoff@mailbox.org>
#
# License: BSD-3-Clause

from pathlib import Path

import mne
import numpy as np
from mne.preprocessing import annotate_amplitude
from mne.utils import logger, verbose
from mne.viz import use_browser_backend

from mne_bids import mark_channels, read_raw_bids
from mne_bids.config import ALLOWED_DATATYPE_EXTENSIONS
from mne_bids.read import _from_tsv, _read_events
from mne_bids.write import _events_tsv


@verbose
def inspect_dataset(
    bids_path,
    find_flat=True,
    l_freq=None,
    h_freq=None,
    show_annotations=True,
    verbose=None,
):
    """Inspect and annotate BIDS raw data.

    This function allows you to browse MEG, EEG, and iEEG raw data stored in a
    BIDS dataset. You can toggle the status of a channel (bad or good) by
    clicking on the traces, and when closing the browse window, you will be
    asked whether you want to save the changes to the existing BIDS dataset or
    discard them.

    .. warning:: This functionality is still experimental and will be extended
                 in the future. Its API will likely change. Planned features
                 include automated bad channel detection and visualization of
                 MRI images.

    .. note:: Currently, only MEG, EEG, and iEEG data can be inspected.

    To add or modify annotations, press ``A`` to toggle annotation mode.

    Parameters
    ----------
    bids_path : BIDSPath
        A :class:`mne_bids.BIDSPath` containing at least a ``root``. All
        matching files will be inspected. To select only a subset of the data,
        set more :class:`mne_bids.BIDSPath` attributes. If ``datatype`` is not
        set and multiple datatypes are found, they will be inspected in the
        following order: MEG, EEG, iEEG.
        To read a specific file, set all the :class:`mne_bids.BIDSPath`
        attributes required to uniquely identify the file: If this ``BIDSPath``
        is accepted by :func:`mne_bids.read_raw_bids`, it will work here.
    find_flat : bool
        Whether to auto-detect channels producing "flat" signals, i.e., with
        unusually low variability. Flat **segments** will be added to
        ``*_events.tsv``, while channels with more than 5 percent of flat data
        will be marked as ``bad`` in ``*_channels.tsv``.

        .. note::
            This function calls ``mne.preprocessing.annotate_amplitude``
            (MNE-Python 1.0 or newer) or ``mne.preprocessing.annotate_flat``
            (older versions of MNE-Python)
            and will only consider segments of at least **50 ms consecutive
            flatness** as "flat" (deviating from MNE-Python's default of 5 ms).
            If more than 5 percent of a channel's data has been marked as flat,
            the entire channel will be added to the list of bad channels. Only
            flat time segments applying to channels **not** marked as bad will
            be added to ``*_events.tsv``.

    l_freq : float | None
        The high-pass filter cutoff frequency to apply when displaying the
        data. This can be useful when inspecting data with slow drifts. If
        ``None``, no high-pass filter will be applied.
    h_freq : float | None
        The low-pass filter cutoff frequency to apply when displaying the
        data. This can be useful when inspecting data with high-frequency
        artifacts. If ``None``, no low-pass filter will be applied.
    show_annotations : bool
        Whether to show annotations (events, bad segments, …) or not. If
        ``False``, toggling annotations mode by pressing ``A`` will be disabled
        as well.
    %(verbose)s

    Examples
    --------
    Disable flat channel & segment detection, and apply a filter with a
    passband of 1–30 Hz.

    >>> from mne_bids import BIDSPath
    >>> root = Path('./mne_bids/tests/data/tiny_bids').absolute()
    >>> bids_path = BIDSPath(subject='01', task='rest', session='eeg',
    ...                      suffix='eeg', extension='.vhdr', root=root)
    >>> inspect_dataset(bids_path=bids_path, find_flat=False,  # doctest: +SKIP
    ...                 l_freq=1, h_freq=30)
    """
    allowed_extensions = set(
        ALLOWED_DATATYPE_EXTENSIONS["meg"]
        + ALLOWED_DATATYPE_EXTENSIONS["eeg"]
        + ALLOWED_DATATYPE_EXTENSIONS["ieeg"]
    )

    bids_paths = [
        p
        for p in bids_path.match(check=True)
        if (p.extension is None or p.extension in allowed_extensions)
        and p.acquisition != "crosstalk"
    ]

    for bids_path_ in bids_paths:
        _inspect_raw(
            bids_path=bids_path_,
            l_freq=l_freq,
            h_freq=h_freq,
            find_flat=find_flat,
            show_annotations=show_annotations,
        )


# XXX This this should probably be refactored into a class attribute someday.
_global_vars = dict(raw_fig=None, dialog_fig=None, mne_close_key=None)


def _inspect_raw(*, bids_path, l_freq, h_freq, find_flat, show_annotations):
    """Raw data inspection."""
    # Delay the import
    import matplotlib
    import matplotlib.pyplot as plt

    extra_params = dict()
    if bids_path.extension == ".fif":
        extra_params["allow_maxshield"] = "yes"
    raw = read_raw_bids(bids_path, extra_params=extra_params, verbose="error")
    old_bads = raw.info["bads"].copy()
    old_annotations = raw.annotations.copy()

    if find_flat:
        raw.load_data()  # Speeds up processing dramatically
        flat_annot, flat_chans = annotate_amplitude(
            raw=raw, flat=0, min_duration=0.05, bad_percent=5
        )
        new_annot = raw.annotations + flat_annot
        raw.set_annotations(new_annot)
        raw.info["bads"] = list(set(raw.info["bads"] + flat_chans))
        del new_annot, flat_annot
    else:
        flat_chans = []

    show_options = bids_path.datatype == "meg"

    with use_browser_backend("matplotlib"):
        fig = raw.plot(
            title=f"{bids_path.root.name}: {bids_path.basename}",
            highpass=l_freq,
            lowpass=h_freq,
            show_options=show_options,
            block=False,
            show=False,
            verbose="warning",
        )

    # Add our own event handlers so that when the MNE Raw Browser is being
    # closed, our dialog box will pop up, asking whether to save changes.
    def _handle_close(event):
        mne_raw_fig = event.canvas.figure
        # Bads alterations are only transferred to `inst` once the figure is
        # closed; Annotation changes are immediately reflected in `inst`
        new_bads = mne_raw_fig.mne.info["bads"].copy()
        new_annotations = mne_raw_fig.mne.inst.annotations.copy()

        if not new_annotations:
            # Ensure it's not an empty list, but an empty set of Annotations.
            new_annotations = mne.Annotations(
                onset=[],
                duration=[],
                description=[],
                orig_time=mne_raw_fig.mne.info["meas_date"],
            )
        _save_raw_if_changed(
            old_bads=old_bads,
            new_bads=new_bads,
            flat_chans=flat_chans,
            old_annotations=old_annotations,
            new_annotations=new_annotations,
            bids_path=bids_path,
        )
        _global_vars["raw_fig"] = None

    def _keypress_callback(event):
        if event.key == _global_vars["mne_close_key"]:
            _handle_close(event)

    fig.canvas.mpl_connect("close_event", _handle_close)
    fig.canvas.mpl_connect("key_press_event", _keypress_callback)

    if not show_annotations:
        # Remove annotations and kill `_toggle_annotation_fig` method, since
        # we cannot directly and easily remove the associated `a` keyboard
        # event callback.
        fig._clear_annotations()
        fig._toggle_annotation_fig = lambda: None
        # Ensure it's not an empty list, but an empty set of Annotations.
        old_annotations = mne.Annotations(
            onset=[], duration=[], description=[], orig_time=raw.info["meas_date"]
        )

    if matplotlib.get_backend().lower() != "agg":
        plt.show(block=True)

    _global_vars["raw_fig"] = fig
    _global_vars["mne_close_key"] = fig.mne.close_key


def _annotations_almost_equal(old_annotations, new_annotations):
    """Allow for a tiny bit of floating point precision loss."""
    if (
        np.array_equal(old_annotations.description, new_annotations.description)
        and np.array_equal(old_annotations.orig_time, new_annotations.orig_time)
        and np.allclose(old_annotations.onset, new_annotations.onset)
        and np.allclose(old_annotations.duration, new_annotations.duration)
    ):
        return True
    else:
        return False


def _save_annotations(*, annotations, bids_path):
    # Attach the new Annotations to our raw data so we can easily convert them
    # to events, which will be stored in the *_events.tsv sidecar.
    extra_params = dict()
    if bids_path.extension == ".fif":
        extra_params["allow_maxshield"] = "yes"

    raw = read_raw_bids(
        bids_path=bids_path, extra_params=extra_params, verbose="warning"
    )
    raw.set_annotations(annotations)
    events, durs, descrs = _read_events(
        events=None, event_id=None, bids_path=bids_path, raw=raw
    )

    # Write sidecar – or remove it if no events are left.
    events_tsv_fname = bids_path.copy().update(suffix="events", extension=".tsv").fpath

    if len(events) > 0:
        _events_tsv(
            events=events,
            durations=durs,
            raw=raw,
            fname=events_tsv_fname,
            trial_type=descrs,
            overwrite=True,
        )
    elif events_tsv_fname.exists():
        logger.info(
            f"No events remaining after interactive inspection, "
            f"removing {events_tsv_fname.name}"
        )
        events_tsv_fname.unlink()


def _save_raw_if_changed(
    *, old_bads, new_bads, flat_chans, old_annotations, new_annotations, bids_path
):
    """Save bad channel selection if it has been changed.

    Parameters
    ----------
    old_bads : list
        The original bad channels.
    new_bads : list
        The updated set of bad channels (i.e. **all** of them, not only the
        changed ones).
    flat_chans : list
        The auto-detected flat channels. This is either an empty list or a
        subset of ``new_bads``.
    old_annotations : mne.Annotations
        The original Annotations.
    new_annotations : mne.Annotations
        The new Annotations.
    """
    assert set(flat_chans).issubset(set(new_bads))

    if set(old_bads) == set(new_bads):
        bads = None
        bad_descriptions = []
    else:
        bads = new_bads
        bad_descriptions = []

        # Generate entries for the `status_description` column.
        channels_tsv_fname = (
            bids_path.copy().update(suffix="channels", extension=".tsv").fpath
        )
        channels_tsv_data = _from_tsv(channels_tsv_fname)

        for ch_name in bads:
            idx = channels_tsv_data["name"].index(ch_name)
            if channels_tsv_data["status"][idx] == "bad":
                # Channel was already marked as bad in the data, so retain
                # existing description.
                description = channels_tsv_data["status_description"][idx]
            elif ch_name in flat_chans:
                description = "Flat channel, auto-detected via MNE-BIDS"
            else:
                # Channel has been manually marked as bad during inspection
                description = "Interactive inspection via MNE-BIDS"

            bad_descriptions.append(description)
            del ch_name, description

        del (
            channels_tsv_data,
            channels_tsv_fname,
        )

    if _annotations_almost_equal(old_annotations, new_annotations):
        annotations = None
    else:
        annotations = new_annotations

    if bads is None and annotations is None:
        # Nothing has changed, so we can just exit.
        return None

    return _save_raw_dialog_box(
        bads=bads,
        bad_descriptions=bad_descriptions,
        annotations=annotations,
        bids_path=bids_path,
    )


def _save_raw_dialog_box(*, bads, bad_descriptions, annotations, bids_path):
    """Display a dialog box asking whether to save the changes."""
    # Delay the imports
    import matplotlib
    import matplotlib.pyplot as plt
    from matplotlib.widgets import Button
    from mne.viz.utils import figure_nobar

    title = "Save changes?"
    message = "You have modified "
    if bads is not None and annotations is None:
        message += "the bad channel selection "
        figsize = (7.5, 2.5)
    elif bads is None and annotations is not None:
        message += "the bad segments selection "
        figsize = (7.5, 2.5)
    else:
        message += "the bad channel and\nannotations selection "
        figsize = (8.5, 3)

    message += (
        f"of\n"
        f"{bids_path.basename}.\n\n"
        f"Would you like to save these changes to the\n"
        f"BIDS dataset?"
    )
    icon_fname = str(Path(__file__).parent / "assets" / "help-128px.png")
    icon = plt.imread(icon_fname)

    fig = figure_nobar(figsize=figsize)
    fig.canvas.manager.set_window_title("MNE-BIDS Inspector")
    fig.suptitle(title, y=0.95, fontsize="xx-large", fontweight="bold")

    gs = fig.add_gridspec(1, 2, width_ratios=(1.5, 5))

    # The dialog box tet.
    ax_msg = fig.add_subplot(gs[0, 1])
    ax_msg.text(
        x=0,
        y=0.8,
        s=message,
        fontsize="large",
        verticalalignment="top",
        horizontalalignment="left",
        multialignment="left",
    )
    ax_msg.axis("off")

    # The help icon.
    ax_icon = fig.add_subplot(gs[0, 0])
    ax_icon.imshow(icon)
    ax_icon.axis("off")

    # Buttons.
    ax_save = fig.add_axes([0.6, 0.05, 0.3, 0.1])
    ax_dont_save = fig.add_axes([0.1, 0.05, 0.3, 0.1])

    save_button = Button(ax=ax_save, label="Save")
    save_button.label.set_fontsize("medium")
    save_button.label.set_fontweight("bold")

    dont_save_button = Button(ax=ax_dont_save, label="Don't save")
    dont_save_button.label.set_fontsize("medium")
    dont_save_button.label.set_fontweight("bold")

    # Store references to keep buttons alive.
    fig.save_button = save_button
    fig.dont_save_button = dont_save_button

    # Define callback functions.
    def _save_callback(event):
        plt.close(event.canvas.figure)  # Close dialog
        _global_vars["dialog_fig"] = None

        if bads is not None:
            _save_bads(bads=bads, descriptions=bad_descriptions, bids_path=bids_path)
        if annotations is not None:
            _save_annotations(annotations=annotations, bids_path=bids_path)

    def _dont_save_callback(event):
        plt.close(event.canvas.figure)  # Close dialog
        _global_vars["dialog_fig"] = None

    def _keypress_callback(event):
        if event.key in ["enter", "return"]:
            _save_callback(event)
        elif event.key == _global_vars["mne_close_key"]:
            _dont_save_callback(event)

    # Connect events to callback functions.
    save_button.on_clicked(_save_callback)
    dont_save_button.on_clicked(_dont_save_callback)
    fig.canvas.mpl_connect("close_event", _dont_save_callback)
    fig.canvas.mpl_connect("key_press_event", _keypress_callback)

    if matplotlib.get_backend().lower() != "agg":
        fig.show()

    _global_vars["dialog_fig"] = fig


def _save_bads(*, bads, descriptions, bids_path):
    """Update the set of channels marked as bad.

    Parameters
    ----------
    bads : list
        The complete list of bad channels.
    descriptions : list
        The values to be written to the `status_description` column.
    """
    # We first make all channels not passed as bad here to be marked as good.
    mark_channels(bids_path=bids_path, ch_names=[], status="good")
    mark_channels(
        bids_path=bids_path, ch_names=bads, status="bad", descriptions=descriptions
    )



"""Write raw files to BIDS format.

example usage:  $ mne_bids raw_to_bids --subject_id sub01 --task rest
--raw data.edf --bids_root new_path

"""

# Authors: Teon Brooks <teon.brooks@gmail.com>
#          Stefan Appelhoff <stefan.appelhoff@mailbox.org>
#
# License: BSD-3-Clause
from itertools import chain, repeat

import mne_bids
from mne_bids import BIDSPath, write_raw_bids
from mne_bids.read import _read_raw


def run():
    """Run the raw_to_bids command."""
    from mne.commands.utils import get_optparser

    parser = get_optparser(
        __file__,
        usage="usage: %prog options args",
        prog_prefix="mne_bids",
        version=mne_bids.__version__,
    )

    parser.add_option(
        "--subject_id",
        dest="subject_id",
        help="subject name in BIDS compatible format (01, 02, etc.)",
    )
    parser.add_option(
        "--task", dest="task", help="name of the task the data is based on"
    )
    parser.add_option("--raw", dest="raw_fname", help="path to the raw MEG file")
    parser.add_option(
        "--bids_root", dest="bids_root", help="The path of the BIDS compatible folder."
    )
    parser.add_option(
        "--session_id", dest="session_id", help="session name in BIDS compatible format"
    )
    parser.add_option("--run", dest="run", help="run number for this dataset")
    parser.add_option(
        "--acq", dest="acq", help="acquisition parameter for this dataset"
    )
    parser.add_option("--events", dest="events", help="events file (events.tsv)")
    parser.add_option(
        "--event_id", dest="event_id", help="event id dict", metavar="eid"
    )
    parser.add_option("--hpi", dest="hpi", help="path to the MEG marker points")
    parser.add_option(
        "--electrode", dest="electrode", help="path to head-native digitizer points"
    )
    parser.add_option("--hsp", dest="hsp", help="path to headshape points")
    parser.add_option("--config", dest="config", help="path to the configuration file")
    parser.add_option(
        "--overwrite",
        dest="overwrite",
        help="whether to overwrite existing data (BOOLEAN)",
    )
    parser.add_option(
        "--line_freq",
        dest="line_freq",
        help="The frequency of the line noise in Hz "
        "(e.g. 50 or 60). If unknown, pass None",
    )

    opt, args = parser.parse_args()

    if len(args) > 0:
        parser.print_help()
        parser.error(f'Do not specify arguments without flags. Found: "{args}".\n')

    if not all([opt.subject_id, opt.task, opt.raw_fname, opt.bids_root]):
        parser.print_help()
        parser.error(
            "Arguments missing. You need to specify at least the"
            "following: --subject_id, --task, --raw, --bids_root."
        )

    bids_path = BIDSPath(
        subject=opt.subject_id,
        session=opt.session_id,
        run=opt.run,
        acquisition=opt.acq,
        task=opt.task,
        root=opt.bids_root,
    )

    allow_maxshield = False
    if opt.raw_fname.endswith(".fif"):
        allow_maxshield = "yes"

    raw = _read_raw(
        opt.raw_fname,
        hpi=opt.hpi,
        electrode=opt.electrode,
        hsp=opt.hsp,
        config_path=opt.config,
        allow_maxshield=allow_maxshield,
    )
    if opt.line_freq is not None:
        line_freq = None if opt.line_freq == "None" else float(opt.line_freq)
        raw.info["line_freq"] = line_freq

    if opt.overwrite is not None:
        truthy = [1, "True", "true"]
        falsy = [0, "False", "false"]
        bool_mapping = dict(chain(zip(truthy, repeat(True)), zip(falsy, repeat(False))))
        opt.overwrite = bool_mapping[opt.overwrite]

    write_raw_bids(
        raw,
        bids_path,
        event_id=opt.event_id,
        events=opt.events,
        overwrite=opt.overwrite,
        verbose=True,
    )


if __name__ == "__main__":
    run()



"""Configure all tests."""

import mne
import pytest


def pytest_configure(config):
    """Configure pytest options."""
    # Fixtures
    config.addinivalue_line("usefixtures", "monkeypatch_mne")


@pytest.fixture(scope="session")
def monkeypatch_mne():
    """Monkeypatch MNE to ensure we have download=False everywhere in tests."""
    mne.datasets.utils._MODULES_TO_ENSURE_DOWNLOAD_IS_FALSE_IN_TESTS = (
        "mne",
        "mne_bids",
    )



"""Read/write BIDS compatible electrode/coords structures from MNE."""

# Authors: Adam Li <adam2392@gmail.com>
#          Stefan Appelhoff <stefan.appelhoff@mailbox.org>
#          Alex Rockhill <aprockhill@mailbox.org>
#
# License: BSD-3-Clause
import json
import os.path as op
import re
import warnings
from collections import OrderedDict
from pathlib import Path

import mne
import numpy as np
from mne.io.constants import FIFF
from mne.io.pick import _picks_to_idx
from mne.utils import _check_option, _validate_type, get_subjects_dir, logger

from mne_bids.config import (
    ALLOWED_SPACES,
    BIDS_COORD_FRAME_DESCRIPTIONS,
    BIDS_COORDINATE_UNITS,
    BIDS_STANDARD_TEMPLATE_COORDINATE_SYSTEMS,
    BIDS_TO_MNE_FRAMES,
    MNE_FRAME_TO_STR,
    MNE_STR_TO_FRAME,
    MNE_TO_BIDS_FRAMES,
)
from mne_bids.path import BIDSPath
from mne_bids.tsv_handler import _from_tsv
from mne_bids.utils import (
    _import_nibabel,
    _scale_coord_to_meters,
    _write_json,
    _write_tsv,
    verbose,
    warn,
)

data_dir = Path(__file__).parent / "data"


def _handle_electrodes_reading(electrodes_fname, coord_frame, coord_unit):
    """Read associated electrodes.tsv and populate raw.

    Handle xyz coordinates and coordinate frame of each channel.
    """
    logger.info("Reading electrode " f"coords from {electrodes_fname}.")
    electrodes_dict = _from_tsv(electrodes_fname)
    ch_names_tsv = electrodes_dict["name"]

    def _float_or_nan(val):
        if val == "n/a":
            return np.nan
        else:
            return float(val)

    # convert coordinates to float and create list of tuples
    electrodes_dict["x"] = [_float_or_nan(x) for x in electrodes_dict["x"]]
    electrodes_dict["y"] = [_float_or_nan(x) for x in electrodes_dict["y"]]
    electrodes_dict["z"] = [_float_or_nan(x) for x in electrodes_dict["z"]]
    ch_names_raw = [
        x for i, x in enumerate(ch_names_tsv) if electrodes_dict["x"][i] != "n/a"
    ]
    ch_locs = np.c_[electrodes_dict["x"], electrodes_dict["y"], electrodes_dict["z"]]

    # convert coordinates to meters
    ch_locs = _scale_coord_to_meters(ch_locs, coord_unit)

    # create mne.DigMontage
    ch_pos = dict(zip(ch_names_raw, ch_locs))
    montage = mne.channels.make_dig_montage(ch_pos=ch_pos, coord_frame=coord_frame)
    return montage


def _handle_coordsystem_reading(coordsystem_fpath, datatype):
    """Read associated coordsystem.json.

    Handle reading the coordinate frame and coordinate unit
    of each electrode.
    """
    with open(coordsystem_fpath, encoding="utf-8-sig") as fin:
        coordsystem_json = json.load(fin)

    if datatype == "meg":
        coord_frame = coordsystem_json["MEGCoordinateSystem"]
        coord_unit = coordsystem_json["MEGCoordinateUnits"]
        coord_frame_desc = coordsystem_json.get("MEGCoordinateDescription", None)
    elif datatype == "eeg":
        coord_frame = coordsystem_json["EEGCoordinateSystem"]
        coord_unit = coordsystem_json["EEGCoordinateUnits"]
        coord_frame_desc = coordsystem_json.get("EEGCoordinateDescription", None)
    elif datatype == "ieeg":
        coord_frame = coordsystem_json["iEEGCoordinateSystem"]
        coord_unit = coordsystem_json["iEEGCoordinateUnits"]
        coord_frame_desc = coordsystem_json.get("iEEGCoordinateDescription", None)

    msg = f"Reading coordinate system frame {coord_frame}"
    if coord_frame_desc:
        msg += f": {coord_frame_desc}"

    return coord_frame, coord_unit


def _get_impedances(raw, names):
    """Get the impedance values in kOhm from raw.impedances."""
    if not hasattr(raw, "impedances"):  # pragma: no cover
        return ["n/a"] * len(names)
    no_info = {"imp": np.nan, "imp_unit": "kOhm"}
    impedance_dicts = [raw.impedances.get(name, no_info) for name in names]
    # If we encounter a unit not defined in `scalings`, return NaN
    scalings = {"kOhm": 1, "Ohm": 0.001}
    impedances = [
        imp_dict["imp"] * scalings.get(imp_dict["imp_unit"], np.nan)
        for imp_dict in impedance_dicts
    ]
    # replace np.nan with BIDS 'n/a' representation
    impedances = [i if not np.isnan(i) else "n/a" for i in impedances]
    return impedances


def _write_electrodes_tsv(raw, fname, datatype, overwrite=False):
    """Create an electrodes.tsv file and save it.

    Parameters
    ----------
    raw : mne.io.Raw
        The data as MNE-Python Raw object.
    fname : str
        Filename to save the electrodes.tsv to.
    datatype : str
        Type of the data recording. Can be ``meg``, ``eeg``,
        or ``ieeg``.
    overwrite : bool
        Defaults to False.
        Whether to overwrite the existing data in the file.
        If there is already data for the given `fname` and overwrite is False,
        an error will be raised.

    """
    # create list of channel coordinates and names
    x, y, z, names = list(), list(), list(), list()
    for ch in raw.info["chs"]:
        if ch["kind"] == FIFF.FIFFV_STIM_CH:
            logger.debug(f"Not writing stim chan {ch['ch_name']} " f"to electrodes.tsv")
            continue
        elif np.isnan(ch["loc"][:3]).any() or np.allclose(ch["loc"][:3], 0):
            x.append("n/a")
            y.append("n/a")
            z.append("n/a")
        else:
            x.append(ch["loc"][0])
            y.append(ch["loc"][1])
            z.append(ch["loc"][2])
        names.append(ch["ch_name"])

    # create OrderedDict to write to tsv file
    if datatype == "ieeg":
        # XXX: size should be included in the future
        sizes = ["n/a"] * len(names)
        data = OrderedDict(
            [
                ("name", names),
                ("x", x),
                ("y", y),
                ("z", z),
                ("size", sizes),
            ]
        )
    elif datatype == "eeg":
        data = OrderedDict(
            [
                ("name", names),
                ("x", x),
                ("y", y),
                ("z", z),
            ]
        )
    else:  # pragma: no cover
        raise RuntimeError(f"datatype {datatype} not supported.")

    # Add impedance values if available, currently only BrainVision:
    # https://github.com/mne-tools/mne-python/pull/7974
    if hasattr(raw, "impedances"):
        data["impedance"] = _get_impedances(raw, names)

    # note that any coordsystem.json file shared within sessions
    # will be the same across all runs (currently). So
    # overwrite is set to True always
    # XXX: improve later when BIDS is updated
    # check that there already exists a coordsystem.json
    if Path(fname).exists() and not overwrite:
        electrodes_tsv = _from_tsv(fname)

        # cast values to str to make equality check work
        if any(
            [
                list(map(str, vals1)) != list(vals2)
                for vals1, vals2 in zip(data.values(), electrodes_tsv.values())
            ]
        ):
            raise RuntimeError(
                f"Trying to write electrodes.tsv, but it already "
                f"exists at {fname} and the contents do not match. "
                f"You must differentiate this electrodes.tsv file "
                f'from the existing one, or set "overwrite" to True.'
            )
    _write_tsv(fname, data, overwrite=True)


def _write_optodes_tsv(raw, fname, overwrite=False, verbose=True):
    """Create a optodes.tsv file and save it.

    Parameters
    ----------
    raw : instance of Raw
        The data as MNE-Python Raw object.
    fname : str | BIDSPath
        Filename to save the optodes.tsv to.
    overwrite : bool
        Whether to overwrite the existing file.
        Defaults to False.
    verbose : bool
        Set verbose output to True or False.
    """
    picks = _picks_to_idx(raw.info, "fnirs", exclude=[], allow_empty=True)
    sources = np.zeros(picks.shape)
    detectors = np.zeros(picks.shape)
    for ii in picks:
        # NIRS channel names take a specific form in MNE-Python.
        # The channel names always reflect the source and detector
        # pair, followed by the wavelength frequency.
        # The following code extracts the source and detector
        # numbers from the channel name.
        ch1_name_info = re.match(r"S(\d+)_D(\d+) (\d+)", raw.info["chs"][ii]["ch_name"])
        sources[ii] = ch1_name_info.groups()[0]
        detectors[ii] = ch1_name_info.groups()[1]
    unique_sources = np.unique(sources)
    n_sources = len(unique_sources)
    unique_detectors = np.unique(detectors)
    names = np.concatenate(
        (
            ["S" + str(s) for s in unique_sources.astype(int)],
            ["D" + str(d) for d in unique_detectors.astype(int)],
        )
    )

    xs = np.zeros(names.shape)
    ys = np.zeros(names.shape)
    zs = np.zeros(names.shape)
    for i, source in enumerate(unique_sources):
        s_idx = np.where(sources == source)[0][0]
        xs[i] = raw.info["chs"][s_idx]["loc"][3]
        ys[i] = raw.info["chs"][s_idx]["loc"][4]
        zs[i] = raw.info["chs"][s_idx]["loc"][5]
    for i, detector in enumerate(unique_detectors):
        d_idx = np.where(detectors == detector)[0][0]
        xs[i + n_sources] = raw.info["chs"][d_idx]["loc"][6]
        ys[i + n_sources] = raw.info["chs"][d_idx]["loc"][7]
        zs[i + n_sources] = raw.info["chs"][d_idx]["loc"][8]

    ch_data = {
        "name": names,
        "type": np.concatenate(
            (
                np.full(len(unique_sources), "source"),
                np.full(len(unique_detectors), "detector"),
            )
        ),
        "x": xs,
        "y": ys,
        "z": zs,
    }
    _write_tsv(fname, ch_data, overwrite, verbose)


def _write_coordsystem_json(
    *,
    raw,
    unit,
    hpi_coord_system,
    sensor_coord_system,
    fname,
    datatype,
    overwrite=False,
):
    """Create a coordsystem.json file and save it.

    Parameters
    ----------
    raw : mne.io.Raw
        The data as MNE-Python Raw object.
    unit : str
        Units to be used in the coordsystem specification,
        as in BIDS_COORDINATE_UNITS.
    hpi_coord_system : str
        Name of the coordinate system for the head coils.
    sensor_coord_system : str | tuple of str
        Name of the coordinate system for the sensor positions.
        If a tuple of strings, should be in the form:
        ``(BIDS coordinate frame, MNE coordinate frame)``.
    fname : str
        Filename to save the coordsystem.json to.
    datatype : str
        Type of the data recording. Can be ``meg``, ``eeg``,
        or ``ieeg``.
    overwrite : bool
        Whether to overwrite the existing file.
        Defaults to False.

    """
    if raw.get_montage() is None:
        dig = list()
        coords = dict()
    else:
        montage = raw.get_montage()
        pos = montage.get_positions()
        dig = list() if montage.dig is None else montage.dig
        coords = dict(
            NAS=list() if pos["nasion"] is None else pos["nasion"].tolist(),
            LPA=list() if pos["lpa"] is None else pos["lpa"].tolist(),
            RPA=list() if pos["rpa"] is None else pos["rpa"].tolist(),
        )

    # get the coordinate frame description
    sensor_coord_system_descr = BIDS_COORD_FRAME_DESCRIPTIONS.get(
        sensor_coord_system.lower(), "n/a"
    )

    # create the coordinate json data structure based on 'datatype'
    if datatype == "meg":
        hpi = {d["ident"]: d for d in dig if d["kind"] == FIFF.FIFFV_POINT_HPI}
        if hpi:
            for ident in hpi.keys():
                coords["coil%d" % ident] = hpi[ident]["r"].tolist()

        fid_json = {
            "MEGCoordinateSystem": sensor_coord_system,
            "MEGCoordinateUnits": unit,  # XXX validate this
            "MEGCoordinateSystemDescription": sensor_coord_system_descr,
            "HeadCoilCoordinates": coords,
            "HeadCoilCoordinateSystem": hpi_coord_system,
            "HeadCoilCoordinateUnits": unit,  # XXX validate this
            "AnatomicalLandmarkCoordinates": coords,
            "AnatomicalLandmarkCoordinateSystem": sensor_coord_system,
            "AnatomicalLandmarkCoordinateUnits": unit,
        }
    elif datatype == "eeg":
        fid_json = {
            "EEGCoordinateSystem": sensor_coord_system,
            "EEGCoordinateUnits": unit,
            "EEGCoordinateSystemDescription": sensor_coord_system_descr,
            "AnatomicalLandmarkCoordinates": coords,
            "AnatomicalLandmarkCoordinateSystem": sensor_coord_system,
            "AnatomicalLandmarkCoordinateUnits": unit,
        }
    elif datatype == "ieeg":
        fid_json = {
            # (Other, Pixels, ACPC)
            "iEEGCoordinateSystem": sensor_coord_system,
            "iEEGCoordinateSystemDescription": sensor_coord_system_descr,
            "iEEGCoordinateUnits": unit,  # m (MNE), mm, cm , or pixels
        }
    elif datatype == "nirs":
        fid_json = {
            "NIRSCoordinateSystem": sensor_coord_system,
            "NIRSCoordinateSystemDescription": sensor_coord_system_descr,
            "NIRSCoordinateUnits": unit,
        }

    # note that any coordsystem.json file shared within sessions
    # will be the same across all runs (currently). So
    # overwrite is set to True always
    # XXX: improve later when BIDS is updated
    # check that there already exists a coordsystem.json
    if Path(fname).exists() and not overwrite:
        with open(fname, encoding="utf-8-sig") as fin:
            coordsystem_dict = json.load(fin)
        if fid_json != coordsystem_dict:
            raise RuntimeError(
                f"Trying to write coordsystem.json, but it already "
                f"exists at {fname} and the contents do not match. "
                f"You must differentiate this coordsystem.json file "
                f'from the existing one, or set "overwrite" to True.'
            )
    _write_json(fname, fid_json, overwrite=True)


def _write_dig_bids(bids_path, raw, montage=None, acpc_aligned=False, overwrite=False):
    """Write BIDS formatted DigMontage from Raw instance.

    Handles coordsystem.json and electrodes.tsv writing
    from DigMontage.

    Parameters
    ----------
    bids_path : BIDSPath
        Path in the BIDS dataset to save the ``electrodes.tsv``
        and ``coordsystem.json`` file for. ``datatype``
        attribute must be ``eeg``, or ``ieeg``. For ``meg``
        data, ``electrodes.tsv`` are not saved.
    raw : mne.io.Raw
        The data as MNE-Python Raw object.
    montage : mne.channels.DigMontage | None
        The montage to use rather than the one in ``raw`` if it
        must be transformed from the "head" coordinate frame.
    acpc_aligned : bool
        Whether "mri" space is aligned to ACPC.
    overwrite : bool
        Whether to overwrite the existing file.
        Defaults to False.

    """
    # write electrodes data for iEEG and EEG
    unit = "m"  # defaults to meters

    if montage is None:
        montage = raw.get_montage()
    else:  # assign montage to raw but supress any coordinate transforms
        montage = montage.copy()  # don't modify original
        montage_coord_frame = montage.get_positions()["coord_frame"]
        fids = [
            d
            for d in montage.dig  # save to add back
            if d["kind"] == FIFF.FIFFV_POINT_CARDINAL
        ]
        montage.remove_fiducials()  # prevent coordinate transform
        with warnings.catch_warnings():
            warnings.filterwarnings(
                action="ignore",
                category=RuntimeWarning,
                message=".*nasion not found",
                module="mne",
            )
            raw.set_montage(montage)
        for ch in raw.info["chs"]:
            ch["coord_frame"] = MNE_STR_TO_FRAME[montage_coord_frame]
        for d in raw.info["dig"]:
            d["coord_frame"] = MNE_STR_TO_FRAME[montage_coord_frame]
        with raw.info._unlock():  # add back fiducials
            raw.info["dig"] = fids + raw.info["dig"]

    # get the accepted mne-python coordinate frames
    coord_frame_int = int(montage.dig[0]["coord_frame"])
    mne_coord_frame = MNE_FRAME_TO_STR.get(coord_frame_int, None)
    coord_frame = MNE_TO_BIDS_FRAMES.get(mne_coord_frame, None)

    if coord_frame == "CapTrak" and bids_path.datatype in ("eeg", "nirs"):
        pos = raw.get_montage().get_positions()
        if any([pos[fid_key] is None for fid_key in ("nasion", "lpa", "rpa")]):
            raise RuntimeError(
                "'head' coordinate frame must contain nasion "
                "and left and right pre-auricular point "
                "landmarks"
            )

    if (
        bids_path.datatype == "ieeg"
        and bids_path.space in (None, "ACPC")
        and mne_coord_frame == "ras"
    ):
        if not acpc_aligned:
            raise RuntimeError(
                "`acpc_aligned` is False, if your T1 is not aligned "
                "to ACPC and the coordinates are in fact in ACPC "
                "space there will be no way to relate the coordinates "
                "to the T1. If the T1 is ACPC-aligned, use "
                "`acpc_aligned=True`"
            )
        coord_frame = "ACPC"

    if bids_path.space is None:  # no space, use MNE coord frame
        if coord_frame is None:  # if no MNE coord frame, skip
            warn(
                "Coordinate frame could not be inferred from the raw object "
                "and the BIDSPath.space was none, skipping the writing of "
                "channel positions"
            )
            return
    else:  # either a space and an MNE coord frame or just a space
        if coord_frame is None:  # just a space, use that
            coord_frame = bids_path.space
        else:  # space and raw have coordinate frame, check match
            if bids_path.space != coord_frame and not (
                coord_frame == "fsaverage" and bids_path.space == "MNI305"
            ):  # fsaverage == MNI305
                raise ValueError(
                    "Coordinates in the raw object or montage "
                    f"are in the {coord_frame} coordinate "
                    "frame but BIDSPath.space is "
                    f"{bids_path.space}"
                )

    # create electrodes/coordsystem files using a subset of entities
    # that are specified for these files in the specification
    coord_file_entities = {
        "root": bids_path.root,
        "datatype": bids_path.datatype,
        "subject": bids_path.subject,
        "session": bids_path.session,
        "acquisition": bids_path.acquisition,
        "space": None if bids_path.datatype == "nirs" else coord_frame,
    }
    channels_suffix = "optodes" if bids_path.datatype == "nirs" else "electrodes"
    _channels_fun = (
        _write_optodes_tsv if bids_path.datatype == "nirs" else _write_electrodes_tsv
    )
    channels_path = BIDSPath(
        **coord_file_entities, suffix=channels_suffix, extension=".tsv"
    )
    coordsystem_path = BIDSPath(
        **coord_file_entities, suffix="coordsystem", extension=".json"
    )

    # Now write the data to the elec coords and the coordsystem
    _channels_fun(raw, channels_path, bids_path.datatype, overwrite)
    _write_coordsystem_json(
        raw=raw,
        unit=unit,
        hpi_coord_system="n/a",
        sensor_coord_system=coord_frame,
        fname=coordsystem_path,
        datatype=bids_path.datatype,
        overwrite=overwrite,
    )


def _read_dig_bids(electrodes_fpath, coordsystem_fpath, datatype, raw):
    """Read MNE-Python formatted DigMontage from BIDS files.

    Handles coordinatesystem.json and electrodes.tsv reading
    to DigMontage.

    Parameters
    ----------
    electrodes_fpath : str
        Filepath of the electrodes.tsv to read.
    coordsystem_fpath : str
        Filepath of the coordsystem.json to read.
    datatype : str
        Type of the data recording. Can be ``meg``, ``eeg``,
        or ``ieeg``.
    raw : mne.io.Raw
        The raw data as MNE-Python ``Raw`` object. The montage
        will be set in place.
    """
    bids_coord_frame, bids_coord_unit = _handle_coordsystem_reading(
        coordsystem_fpath, datatype
    )

    if bids_coord_frame not in ALLOWED_SPACES[datatype]:
        warn(
            f'"{bids_coord_frame}" is not a BIDS-acceptable coordinate frame '
            f"for {datatype.upper()}. The supported coordinate frames are: "
            "{}".format(ALLOWED_SPACES[datatype])
        )
        coord_frame = None
    elif bids_coord_frame in BIDS_TO_MNE_FRAMES:
        coord_frame = BIDS_TO_MNE_FRAMES.get(bids_coord_frame, None)
    else:
        warn(
            f"{bids_coord_frame} is not an MNE-Python coordinate frame "
            f"for {datatype.upper()} data and so will be set to 'unknown'"
        )
        coord_frame = "unknown"

    # check coordinate units
    if bids_coord_unit not in BIDS_COORDINATE_UNITS:
        warn(
            f"Coordinate unit is not an accepted BIDS unit for "
            f"{electrodes_fpath}. Please specify to be one of "
            f"{BIDS_COORDINATE_UNITS}. Skipping electrodes.tsv reading..."
        )
        coord_frame = None

    # montage is interpretable only if coordinate frame was properly parsed
    if coord_frame is not None:
        # read in electrode coordinates as a DigMontage object
        montage = _handle_electrodes_reading(
            electrodes_fpath, coord_frame, bids_coord_unit
        )
    else:
        montage = None

    if montage is not None:
        # determine if there are problematic channels
        ch_pos = montage._get_ch_pos()
        nan_chs = []
        for ch_name, ch_coord in ch_pos.items():
            if any(np.isnan(ch_coord)) and ch_name not in raw.info["bads"]:
                nan_chs.append(ch_name)
        if len(nan_chs) > 0:
            warn(
                f"There are channels without locations "
                f"(n/a) that are not marked as bad: {nan_chs}"
            )

    # add montage to Raw object
    # XXX: Starting with mne 0.24, this will raise a RuntimeWarning
    #      if channel types are included outside of
    #      (EEG/sEEG/ECoG/DBS/fNIRS). Probably needs a fix in the future.
    with warnings.catch_warnings():
        warnings.filterwarnings(
            action="ignore",
            category=RuntimeWarning,
            message=".*nasion not found",
            module="mne",
        )
        raw.set_montage(montage, on_missing="warn")

    # put back in unknown for unknown coordinate frame
    if coord_frame == "unknown":
        for ch in raw.info["chs"]:
            ch["coord_frame"] = MNE_STR_TO_FRAME["unknown"]
        for d in raw.info["dig"]:
            d["coord_frame"] = MNE_STR_TO_FRAME["unknown"]


@verbose
def template_to_head(info, space, coord_frame="auto", unit="auto", verbose=None):
    """Transform a BIDS standard template montage to the head coordinate frame.

    Parameters
    ----------
    %(info_not_none)s The info is modified in place.
    space : str
        The name of the BIDS standard template. See
        https://bids-specification.readthedocs.io/en/latest/appendices/coordinate-systems.html#standard-template-identifiers
        for a list of acceptable spaces.
    coord_frame : 'mri' | 'mri_voxel' | 'ras'
        BIDS template coordinate systems do not specify a coordinate frame,
        so this must be determined by inspecting the documentation for the
        dataset or the ``electrodes.tsv`` file.  If ``'auto'``, the coordinate
        frame is assumed to be ``'mri_voxel'`` if the coordinates are strictly
        positive, and ``'ras'`` (``"scanner RAS"``) otherwise.

        .. warning::

            ``scanner RAS`` and ``surface RAS`` coordinates frames are similar
            so be very careful not to assume a BIDS dataset's coordinates are
            in one when they are actually in the other. The only way to tell
            for template coordinate systems, currently, is if it is specified
            in the dataset documentation.

    unit : 'm' | 'mm' | 'auto'
        The unit that was used in the coordinate system specification.
        If ``'auto'``, ``'m'`` will be inferred if the montage
        spans less than ``-1`` to ``1``, and ``'mm'`` otherwise. If the
        ``coord_frame`` is ``'mri_voxel'``, ``unit`` will be ignored.
    %(verbose)s

    Returns
    -------
    %(info_not_none)s The modified ``Info`` object.
    trans : mne.transforms.Transform
        The data transformation matrix from ``'head'`` to ``'mri'``
        coordinates.

    """
    _validate_type(info, mne.Info)
    _check_option("space", space, BIDS_STANDARD_TEMPLATE_COORDINATE_SYSTEMS)
    _check_option("coord_frame", coord_frame, ("auto", "mri", "mri_voxel", "ras"))
    _check_option("unit", unit, ("auto", "m", "mm"))
    montage = info.get_montage()
    if montage is None:
        raise RuntimeError("No montage found in the `raw` object")
    montage.remove_fiducials()  # we will add fiducials so remove any
    pos = montage.get_positions()
    if pos["coord_frame"] not in ("mni_tal", "unknown"):
        raise RuntimeError(
            "Montage coordinate frame '{}' not expected for a template "
            "montage, should be 'unknown' or 'mni_tal'".format(pos["coord_frame"])
        )
    locs = np.array(list(pos["ch_pos"].values()))
    locs = locs[~np.any(np.isnan(locs), axis=1)]  # only channels with loc
    if locs.size == 0:
        raise RuntimeError("No channel locations found in the montage")
    if unit == "auto":
        unit = "m" if abs(locs - locs.mean(axis=0)).max() < 1 else "mm"
    if coord_frame == "auto":
        coord_frame = "mri_voxel" if locs.min() >= 0 else "ras"
    # transform montage to head
    # set to the right coordinate frame as specified by the user
    for d in montage.dig:  # ensure same coordinate frame
        d["coord_frame"] = MNE_STR_TO_FRAME[coord_frame]
    # do the transforms, first ras -> vox if needed
    if montage.get_positions()["coord_frame"] == "ras":
        ras_vox_trans = mne.read_trans(data_dir / f"space-{space}_ras-vox_trans.fif")
        if unit == "m":  # must be in mm here
            for d in montage.dig:
                d["r"] *= 1000
        montage.apply_trans(ras_vox_trans)
    if montage.get_positions()["coord_frame"] == "mri_voxel":
        vox_mri_trans = mne.read_trans(data_dir / f"space-{space}_vox-mri_trans.fif")
        montage.apply_trans(vox_mri_trans)
    assert montage.get_positions()["coord_frame"] == "mri"
    if not (unit == "m" and coord_frame == "mri"):  # if so, already in m
        for d in montage.dig:
            d["r"] /= 1000  # mm -> m
    # now add fiducials (in mri coordinates)
    fids = mne.io.read_fiducials(data_dir / f"space-{space}_fiducials.fif")[0]
    montage.dig = fids + montage.dig  # add fiducials
    for fid in fids:  # ensure also in mri
        fid["coord_frame"] = MNE_STR_TO_FRAME["mri"]
    info.set_montage(montage)  # transform to head
    # finally return montage
    return info, mne.read_trans(data_dir / f"space-{space}_trans.fif")


@verbose
def convert_montage_to_ras(montage, subject, subjects_dir=None, verbose=None):
    """Convert a montage from surface RAS (m) to scanner RAS (m).

    Parameters
    ----------
    montage : mne.channels.DigMontage
        The montage in the "mri" coordinate frame. Note: modified in place.
    %(subject)s
    %(subjects_dir)s
    %(verbose)s
    """
    nib = _import_nibabel("converting a montage to RAS")

    subjects_dir = get_subjects_dir(subjects_dir, raise_error=True)
    T1_fname = op.join(subjects_dir, subject, "mri", "T1.mgz")
    if not op.isfile(T1_fname):
        raise RuntimeError(
            f"Freesurfer subject ({subject}) and/or "
            f"subjects_dir ({subjects_dir}, incorrectly "
            "formatted, T1.mgz not found"
        )
    T1 = nib.load(T1_fname)

    # transform from "mri" (Freesurfer surface RAS) to "ras" (scanner RAS)
    mri_vox_t = np.linalg.inv(T1.header.get_vox2ras_tkr())
    mri_vox_t[:3, :3] *= 1000  # scale from mm to m
    mri_vox_trans = mne.transforms.Transform(fro="mri", to="mri_voxel", trans=mri_vox_t)

    vox_ras_t = T1.header.get_vox2ras()
    vox_ras_t[:3] /= 1000  # scale from mm to m
    vox_ras_trans = mne.transforms.Transform(fro="mri_voxel", to="ras", trans=vox_ras_t)
    montage.apply_trans(  # mri->vox + vox->ras = mri->ras
        mne.transforms.combine_transforms(
            mri_vox_trans, vox_ras_trans, fro="mri", to="ras"
        )
    )


@verbose
def convert_montage_to_mri(montage, subject, subjects_dir=None, verbose=None):
    """Convert a montage from scanner RAS (m) to surface RAS (m).

    Parameters
    ----------
    montage : mne.channels.DigMontage
        The montage in the "ras" coordinate frame. Note: modified in place.
    %(subject)s
    %(subjects_dir)s
    %(verbose)s

    Returns
    -------
    ras_mri_t : mne.transforms.Transform
        The transformation matrix from ``'ras'`` (``scanner RAS``) to
        ``'mri'`` (``surface RAS``).
    """
    nib = _import_nibabel("converting a montage to MRI")

    subjects_dir = get_subjects_dir(subjects_dir, raise_error=True)
    T1_fname = op.join(subjects_dir, subject, "mri", "T1.mgz")
    if not op.isfile(T1_fname):
        raise RuntimeError(
            f"Freesurfer subject ({subject}) and/or "
            f"subjects_dir ({subjects_dir}, incorrectly "
            "formatted, T1.mgz not found"
        )
    T1 = nib.load(T1_fname)

    # transform from "ras" (scanner RAS) to "mri" (Freesurfer surface RAS)
    ras_vox_t = T1.header.get_ras2vox()
    ras_vox_t[:3, :3] *= 1000  # scale from mm to m
    ras_vox_trans = mne.transforms.Transform(fro="ras", to="mri_voxel", trans=ras_vox_t)

    vox_mri_t = T1.header.get_vox2ras_tkr()
    vox_mri_t[:3] /= 1000  # scale from mm to m
    vox_mri_trans = mne.transforms.Transform(fro="mri_voxel", to="mri", trans=vox_mri_t)
    montage.apply_trans(  # ras->vox + vox->mri = ras->mri
        mne.transforms.combine_transforms(
            ras_vox_trans, vox_mri_trans, fro="ras", to="mri"
        )
    )



"""Mark channels in an existing BIDS dataset as "bad".

example usage:
$ mne_bids mark_channels --ch_name="MEG 0112" --description="noisy" \
                         --ch_name="MEG 0131" --description="flat" \
                         --subject_id=01 --task=experiment --session=test \
                         --bids_root=bids_root

"""
# Authors: Richard Höchenberger <richard.hoechenberger@gmail.com>
#
# License: BSD-3-Clause

from mne.utils import logger

import mne_bids
from mne_bids import BIDSPath, mark_channels
from mne_bids.config import reader


def run():
    """Run the mark_channels command."""
    from mne.commands.utils import get_optparser

    parser = get_optparser(
        __file__,
        usage="usage: %prog options args",
        prog_prefix="mne_bids",
        version=mne_bids.__version__,
    )

    parser.add_option(
        "--ch_name",
        dest="ch_names",
        action="append",
        default=[],
        help="The names of the bad channels. If multiple "
        "channels are bad, pass the --ch_name parameter "
        "multiple times.",
    )
    parser.add_option(
        "--status",
        default="bad",
        help='Status of the channels (Either "good", or "bad").',
    )
    parser.add_option(
        "--description",
        dest="descriptions",
        action="append",
        default=[],
        help="Descriptions as to why the channels are bad. "
        "Must match the number of bad channels provided. "
        "Pass multiple times to supply more than one "
        "value in that case.",
    )
    parser.add_option(
        "--bids_root",
        dest="bids_root",
        help="The path of the folder containing the BIDS dataset",
    )
    parser.add_option("--subject_id", dest="subject", help="Subject name")
    parser.add_option("--session_id", dest="session", help="Session name")
    parser.add_option("--task", dest="task", help="Task name")
    parser.add_option("--acq", dest="acquisition", help="Acquisition parameter")
    parser.add_option("--run", dest="run", help="Run number")
    parser.add_option("--proc", dest="processing", help="Processing label.")
    parser.add_option("--rec", dest="recording", help="Recording name")
    parser.add_option(
        "--type", dest="datatype", help="Recording data type, e.g. meg, ieeg or eeg"
    )
    parser.add_option(
        "--suffix",
        dest="suffix",
        help="The filename suffix, i.e. the last part before the extension",
    )
    parser.add_option(
        "--ext",
        dest="extension",
        help="The filename extension, including the leading period, e.g. .fif",
    )
    parser.add_option(
        "--verbose",
        dest="verbose",
        action="store_true",
        help="Whether do generate additional diagnostic output",
    )

    opt, args = parser.parse_args()
    if args:
        parser.print_help()
        parser.error(f"Please do not specify arguments without flags. Got: {args}.\n")

    if opt.bids_root is None:
        parser.print_help()
        parser.error("You must specify bids_root")
    if opt.ch_names is None:
        parser.print_help()
        parser.error("You must specify some --ch_name parameters.")

    status = opt.status
    ch_names = [] if opt.ch_names == [""] else opt.ch_names
    bids_path = BIDSPath(
        subject=opt.subject,
        session=opt.session,
        task=opt.task,
        acquisition=opt.acquisition,
        run=opt.run,
        processing=opt.processing,
        recording=opt.recording,
        datatype=opt.datatype,
        suffix=opt.suffix,
        extension=opt.extension,
        root=opt.bids_root,
    )

    bids_paths = bids_path.match()
    # Only keep data we can actually read & write.
    allowed_extensions = list(reader.keys())
    bids_paths = [p for p in bids_paths if p.extension in allowed_extensions]

    if not bids_paths:
        logger.info(
            "No matching files found. Please consider using a less "
            "restrictive set of entities to broaden the search."
        )
        return  # XXX should be return with an error code?

    logger.info(
        f'Marking channels {", ".join(ch_names)} as bad in '
        f"{len(bids_paths)} recording(s) …"
    )
    for bids_path in bids_paths:
        logger.info(f"Processing: {bids_path.basename}")
        mark_channels(
            bids_path=bids_path,
            ch_names=ch_names,
            status=status,
            descriptions=opt.descriptions,
            verbose=opt.verbose,
        )


if __name__ == "__main__":
    run()



"""Configuration values for MNE-BIDS."""

from mne import io
from mne.io.constants import FIFF

BIDS_VERSION = "1.7.0"

PYBV_VERSION = "0.7.3"
EEGLABIO_VERSION = "0.0.2"

DOI = """https://doi.org/10.21105/joss.01896"""

EPHY_ALLOWED_DATATYPES = ["meg", "eeg", "ieeg", "nirs"]

ALLOWED_DATATYPES = EPHY_ALLOWED_DATATYPES + ["anat", "beh"]

MEG_CONVERT_FORMATS = ["FIF", "auto"]
EEG_CONVERT_FORMATS = ["BrainVision", "auto"]
IEEG_CONVERT_FORMATS = ["BrainVision", "auto"]
NIRS_CONVERT_FORMATS = ["auto"]
CONVERT_FORMATS = {
    "meg": MEG_CONVERT_FORMATS,
    "eeg": EEG_CONVERT_FORMATS,
    "ieeg": IEEG_CONVERT_FORMATS,
    "nirs": NIRS_CONVERT_FORMATS,
}

# Orientation of the coordinate system dependent on manufacturer
ORIENTATION = {
    ".con": "KitYokogawa",
    ".ds": "CTF",
    ".fif": "ElektaNeuromag",
    ".pdf": "4DBti",
    ".sqd": "KitYokogawa",
}

EXT_TO_UNIT_MAP = {".con": "m", ".ds": "cm", ".fif": "m", ".pdf": "m", ".sqd": "m"}

UNITS_MNE_TO_BIDS_MAP = {
    "C": "oC",  # temperature in deg. C
}

meg_manufacturers = {
    ".con": "KIT/Yokogawa",
    ".ds": "CTF",
    ".fif": "Elekta",
    ".meg4": "CTF",
    ".pdf": "4D Magnes",
    ".sqd": "KIT/Yokogawa",
}

eeg_manufacturers = {
    ".vhdr": "Brain Products",
    ".eeg": "Brain Products",
    ".edf": "n/a",
    ".EDF": "n/a",
    ".bdf": "Biosemi",
    ".BDF": "Biosemi",
    ".set": "n/a",
    ".fdt": "n/a",
    ".lay": "Persyst",
    ".dat": "Persyst",
    ".EEG": "Nihon Kohden",
    ".cnt": "Neuroscan",
    ".CNT": "Neuroscan",
    ".bin": "EGI",
    ".cdt": "Curry",
}

ieeg_manufacturers = {
    ".vhdr": "Brain Products",
    ".eeg": "Brain Products",
    ".edf": "n/a",
    ".EDF": "n/a",
    ".set": "n/a",
    ".fdt": "n/a",
    ".mef": "n/a",
    ".nwb": "n/a",
    ".lay": "Persyst",
    ".dat": "Persyst",
    ".EEG": "Nihon Kohden",
}

nirs_manufacturers = {".snirf": "SNIRF"}

# file-extension map to mne-python readers
reader = {
    ".con": io.read_raw_kit,
    ".sqd": io.read_raw_kit,
    ".fif": io.read_raw_fif,
    ".pdf": io.read_raw_bti,
    ".ds": io.read_raw_ctf,
    ".vhdr": io.read_raw_brainvision,
    ".edf": io.read_raw_edf,
    ".EDF": io.read_raw_edf,
    ".bdf": io.read_raw_bdf,
    ".set": io.read_raw_eeglab,
    ".lay": io.read_raw_persyst,
    ".EEG": io.read_raw_nihon,
    ".cnt": io.read_raw_cnt,
    ".CNT": io.read_raw_cnt,
    ".bin": io.read_raw_egi,
    ".snirf": io.read_raw_snirf,
    ".cdt": io.read_raw_curry,
}


# Merge the manufacturer dictionaries in a python2 / python3 compatible way
# MANUFACTURERS dictionary only includes the extension of the input filename
# that mne-python accepts (e.g. BrainVision has three files, but the reader
# takes the filename for `.vhdr`)
MANUFACTURERS = dict()
MANUFACTURERS.update(meg_manufacturers)
MANUFACTURERS.update(eeg_manufacturers)
MANUFACTURERS.update(ieeg_manufacturers)
MANUFACTURERS.update(nirs_manufacturers)

# List of synthetic channels by manufacturer that are to be excluded from the
# channel list. Currently this is only for stimulus channels.
IGNORED_CHANNELS = {
    "KIT/Yokogawa": ["STI 014"],
    "BrainProducts": ["STI 014"],
    "n/a": ["STI 014"],  # for unknown manufacturers, ignore it
    "Biosemi": ["STI 014"],
}

allowed_extensions_meg = [".con", ".sqd", ".fif", ".pdf", ".ds"]
allowed_extensions_eeg = [
    ".vhdr",  # BrainVision, accompanied by .vmrk, .eeg
    ".edf",  # European Data Format
    ".bdf",  # Biosemi
    ".set",  # EEGLAB, potentially accompanied by .fdt
]

allowed_extensions_ieeg = [
    ".vhdr",  # BrainVision, accompanied by .vmrk, .eeg
    ".edf",  # European Data Format
    ".set",  # EEGLAB, potentially accompanied by .fdt
    ".mef",  # MEF: Multiscale Electrophysiology File
    ".nwb",  # Neurodata without borders
]

allowed_extensions_nirs = [
    ".snirf",  # SNIRF
]

# allowed extensions (data formats) in BIDS spec
ALLOWED_DATATYPE_EXTENSIONS = {
    "meg": allowed_extensions_meg,
    "eeg": allowed_extensions_eeg,
    "ieeg": allowed_extensions_ieeg,
    "nirs": allowed_extensions_nirs,
}

# allow additional extensions that are not BIDS
# compliant, but we will convert to the
# recommended formats
ALLOWED_INPUT_EXTENSIONS = (
    allowed_extensions_meg
    + allowed_extensions_eeg
    + allowed_extensions_ieeg
    + allowed_extensions_nirs
    + [".lay", ".EEG", ".cnt", ".CNT", ".bin", ".cdt"]
)

# allowed suffixes (i.e. last "_" delimiter in the BIDS filenames before
# the extension)
ALLOWED_FILENAME_SUFFIX = [
    "meg",
    "markers",
    "eeg",
    "ieeg",
    "T1w",
    "FLASH",  # datatype
    "participants",
    "scans",
    "sessions",
    "electrodes",
    "optodes",
    "channels",
    "coordsystem",
    "events",  # sidecars
    "headshape",
    "digitizer",  # meg-specific sidecars
    "beh",
    "physio",
    "stim",  # behavioral
    "nirs",
]

# converts suffix to known path modalities
SUFFIX_TO_DATATYPE = {
    "meg": "meg",
    "headshape": "meg",
    "digitizer": "meg",
    "markers": "meg",
    "eeg": "eeg",
    "ieeg": "ieeg",
    "T1w": "anat",
    "FLASH": "anat",
}

# allowed BIDS extensions (extension in the BIDS filename)
ALLOWED_FILENAME_EXTENSIONS = (
    ALLOWED_INPUT_EXTENSIONS
    + [".json", ".tsv", ".tsv.gz", ".nii", ".nii.gz"]
    + [".pos", ".eeg", ".vmrk"]
    + [".dat", ".EEG"]  # extra datatype-specific metadata files.
    + [".mrk"]  # extra eeg extensions  # KIT/Yokogawa/Ricoh marker coil
)

# allowed BIDSPath entities
ALLOWED_PATH_ENTITIES = (
    "subject",
    "session",
    "task",
    "run",
    "processing",
    "recording",
    "space",
    "acquisition",
    "split",
    "description",
    "suffix",
    "extension",
)
ALLOWED_PATH_ENTITIES_SHORT = {
    "sub": "subject",
    "ses": "session",
    "task": "task",
    "acq": "acquisition",
    "run": "run",
    "proc": "processing",
    "space": "space",
    "rec": "recording",
    "split": "split",
    "desc": "description",
}

# Annotations to never remove during reading or writing
ANNOTATIONS_TO_KEEP = ("BAD_ACQ_SKIP",)

BIDS_STANDARD_TEMPLATE_COORDINATE_SYSTEMS = [
    "ICBM452AirSpace",
    "ICBM452Warp5Space",
    "IXI549Space",
    "fsaverage",
    "fsaverageSym",
    "fsLR",
    "MNIColin27",
    "MNI152Lin",
    "MNI152NLin2009aSym",
    "MNI152NLin2009bSym",
    "MNI152NLin2009cSym",
    "MNI152NLin2009aAsym",
    "MNI152NLin2009bAsym",
    "MNI152NLin2009cAsym",
    "MNI152NLin6Sym",
    "MNI152NLin6ASym",
    "MNI305",
    "NIHPD",
    "OASIS30AntsOASISAnts",
    "OASIS30Atropos",
    "Talairach",
    "UNCInfant",
]

coordsys_standard_template_deprecated = [
    "fsaverage3",
    "fsaverage4",
    "fsaverage5",
    "fsaverage6",
    "fsaveragesym",
    "UNCInfant0V21",
    "UNCInfant1V21",
    "UNCInfant2V21",
    "UNCInfant0V22",
    "UNCInfant1V22",
    "UNCInfant2V22",
    "UNCInfant0V23",
    "UNCInfant1V23",
    "UNCInfant2V23",
]

# accepted BIDS formats, which may be subject to change
# depending on the specification
BIDS_IEEG_COORDINATE_FRAMES = ["ACPC", "Pixels"]
BIDS_MEG_COORDINATE_FRAMES = [
    "CTF",
    "ElektaNeuromag",
    "4DBti",
    "KitYokogawa",
    "ChietiItab",
]
BIDS_EEG_COORDINATE_FRAMES = ["CapTrak"]

# accepted coordinate SI units
BIDS_COORDINATE_UNITS = ["m", "cm", "mm"]
coordsys_wildcard = ["Other"]
BIDS_SHARED_COORDINATE_FRAMES = (
    BIDS_STANDARD_TEMPLATE_COORDINATE_SYSTEMS
    + coordsys_standard_template_deprecated
    + coordsys_wildcard
)

ALLOWED_SPACES = dict()
ALLOWED_SPACES["meg"] = ALLOWED_SPACES["eeg"] = (
    BIDS_SHARED_COORDINATE_FRAMES
    + BIDS_MEG_COORDINATE_FRAMES
    + BIDS_EEG_COORDINATE_FRAMES
)
ALLOWED_SPACES["ieeg"] = BIDS_SHARED_COORDINATE_FRAMES + BIDS_IEEG_COORDINATE_FRAMES
ALLOWED_SPACES["anat"] = None
ALLOWED_SPACES["beh"] = None

# See: https://bids-specification.readthedocs.io/en/latest/appendices/entity-table.html#encephalography-eeg-ieeg-and-meg  # noqa
ENTITY_VALUE_TYPE = {
    "subject": "label",
    "session": "label",
    "task": "label",
    "run": "index",
    "processing": "label",
    "recording": "label",
    "space": "label",
    "acquisition": "label",
    "split": "index",
    "description": "label",
    "suffix": "label",
    "extension": "label",
}

# mapping from supported BIDs coordinate frames -> MNE
BIDS_TO_MNE_FRAMES = {
    "CTF": "ctf_head",
    "4DBti": "ctf_head",
    "KitYokogawa": "ctf_head",
    "ElektaNeuromag": "head",
    "ChietiItab": "head",
    "CapTrak": "head",
    "ACPC": "ras",  # assumes T1 is ACPC-aligned, if not the coordinates are lost  # noqa
    "fsaverage": "mni_tal",  # XXX: note fsaverage and MNI305 are the same  # noqa
    "MNI305": "mni_tal",
}

# mapping from supported MNE coordinate frames -> BIDS
# XXX: note that there are a lot fewer MNE available coordinate
# systems so the range of BIDS supported coordinate systems we
# can write is limited.
MNE_TO_BIDS_FRAMES = {"ctf_head": "CTF", "head": "CapTrak", "mni_tal": "fsaverage"}

# these coordinate frames in mne-python are related to scalp/meg
# 'meg', 'ctf_head', 'ctf_meg', 'head', 'unknown'
# copied from "mne.transforms.MNE_STR_TO_FRAME"
MNE_STR_TO_FRAME = dict(
    meg=FIFF.FIFFV_COORD_DEVICE,
    mri=FIFF.FIFFV_COORD_MRI,
    mri_voxel=FIFF.FIFFV_MNE_COORD_MRI_VOXEL,
    head=FIFF.FIFFV_COORD_HEAD,
    mni_tal=FIFF.FIFFV_MNE_COORD_MNI_TAL,
    ras=FIFF.FIFFV_MNE_COORD_RAS,
    fs_tal=FIFF.FIFFV_MNE_COORD_FS_TAL,
    ctf_head=FIFF.FIFFV_MNE_COORD_CTF_HEAD,
    ctf_meg=FIFF.FIFFV_MNE_COORD_CTF_DEVICE,
    unknown=FIFF.FIFFV_COORD_UNKNOWN,
)
MNE_FRAME_TO_STR = {val: key for key, val in MNE_STR_TO_FRAME.items()}

# see BIDS specification for description we copied over from each
BIDS_COORD_FRAME_DESCRIPTIONS = {
    "acpc": "The origin of the coordinate system is at the Anterior "
    "Commissure and the negative y-axis is passing through the "
    "Posterior Commissure. The positive z-axis is passing through "
    "a mid-hemispheric point in the superior direction.",
    "pixels": "If electrodes are localized in 2D space (only x and y are "
    "specified and z is n/a), then the positions in this file "
    "must correspond to the locations expressed in pixels on "
    "the photo/drawing/rendering of the electrodes on the brain. "
    "In this case, coordinates must be (row,column) pairs, with "
    "(0,0) corresponding to the upper left pixel and (N,0) "
    "corresponding to the lower left pixel.",
    "ctf": "ALS orientation and the origin between the ears",
    "elektaneuromag": "RAS orientation and the origin between the ears",
    "4dbti": "ALS orientation and the origin between the ears",
    "kityokogawa": "ALS orientation and the origin between the ears",
    "chietiitab": "RAS orientation and the origin between the ears",
    "captrak": (
        "The X-axis goes from the left preauricular point (LPA) through "
        "the right preauricular point (RPA). "
        "The Y-axis goes orthogonally to the X-axis through the nasion (NAS). "
        "The Z-axis goes orthogonally to the XY-plane through the vertex of "
        "the head. "
        'This corresponds to a "RAS" orientation with the origin of the '
        "coordinate system approximately between the ears. "
        "See Appendix VIII in the BIDS specification."
    ),
    "fsaverage": "Defined by FreeSurfer, the MRI (surface RAS) origin is "
    "at the center of a 256×256×256 mm^3 anisotropic volume "
    "(may not be in the center of the head).",
    "icbm452airspace": 'Reference space defined by the "average of 452 '
    'T1-weighted MRIs of normal young adult brains" '
    'with "linear transforms of the subjects into the '
    "atlas space using a 12-parameter affine "
    'transformation"',
    "icbm452warp5space": 'Reference space defined by the "average of 452 '
    'T1-weighted MRIs of normal young adult brains" '
    '"based on a 5th order polynomial transformation '
    'into the atlas space"',
    "ixi549space": 'Reference space defined by the average of the "549 (...) '
    'subjects from the IXI dataset" linearly transformed to '
    "ICBM MNI 452.",
    "fsaveragesym": "The fsaverage is a dual template providing both "
    "volumetric and surface coordinates references. The "
    "volumetric template corresponds to a FreeSurfer variant "
    "of MNI305 space. The fsaverageSym atlas also defines a "
    "symmetric surface reference system (formerly described "
    "as fsaveragesym).",
    "fslr": "The fsLR is a dual template providing both volumetric and "
    "surface coordinates references. The volumetric template "
    "corresponds to MNI152NLin6Asym. Surface templates are given "
    "at several sampling densities: 164k (used by HCP pipelines "
    "for 3T and 7T anatomical analysis), 59k (used by HCP pipelines "
    "for 7T MRI bold and DWI analysis), 32k (used by HCP pipelines "
    "for 3T MRI bold and DWI analysis), or 4k (used by HCP "
    "pipelines for MEG analysis) fsaverage_LR surface "
    "reconstructed from the T1w image.",
    "mnicolin27": "Average of 27 T1 scans of a single subject.",
    "mni152lin": "Also known as ICBM (version with linear coregistration).",
    "mni152nlin6sym": "Also known as symmetric ICBM 6th generation "
    "(non-linear coregistration).",
    "mni152nlin6asym": "A variation of MNI152NLin6Sym built by A. Janke that "
    "is released as the MNI template of FSL. Volumetric "
    "templates included with HCP-Pipelines correspond to "
    "this template too.",
    "mni305": "Also known as avg305.",
    "nihpd": "Pediatric templates generated from the NIHPD sample. Available "
    "for different age groups (4.5–18.5 y.o., 4.5–8.5 y.o., "
    "7–11 y.o., 7.5–13.5 y.o., 10–14 y.o., 13–18.5 y.o. This "
    "template also comes in either -symmetric or -asymmetric flavor.",
    "oasis30antsoasisants": "See https://figshare.com/articles/ANTs_ANTsR_Brain_Templates/915436",  # noqa: E501
    "oasis30atropos": "See https://mindboggle.info/data.html",
    "talairach": "Piecewise linear scaling of the brain is implemented as "
    "described in TT88.",
    "uncinfant": "Infant Brain Atlases from Neonates to 1- and 2-year-olds.",
}

for letter in ("a", "b", "c"):
    for sym in ("Sym", "Asym"):
        BIDS_COORD_FRAME_DESCRIPTIONS[f"mni152nlin2009{letter}{sym}"] = (
            "Also known as ICBM (non-linear coregistration with 40 iterations,"
        )
        " released in 2009). It comes in either three different flavours "
        "each in symmetric or asymmetric version."

REFERENCES = {
    "mne-bids": "Appelhoff, S., Sanderson, M., Brooks, T., Vliet, M., "
    "Quentin, R., Holdgraf, C., Chaumon, M., Mikulan, E., "
    "Tavabi, K., Höchenberger, R., Welke, D., Brunner, C., "
    "Rockhill, A., Larson, E., Gramfort, A. and Jas, M. (2019). "
    "MNE-BIDS: Organizing electrophysiological data into the "
    "BIDS format and facilitating their analysis. Journal of "
    "Open Source Software 4: (1896)."
    "https://doi.org/10.21105/joss.01896",
    "meg": "Niso, G., Gorgolewski, K. J., Bock, E., Brooks, T. L., "
    "Flandin, G., Gramfort, A., Henson, R. N., Jas, M., Litvak, "
    "V., Moreau, J., Oostenveld, R., Schoffelen, J., Tadel, F., "
    "Wexler, J., Baillet, S. (2018). MEG-BIDS, the brain "
    "imaging data structure extended to magnetoencephalography. "
    "Scientific Data, 5, 180110."
    "https://doi.org/10.1038/sdata.2018.110",
    "eeg": "Pernet, C. R., Appelhoff, S., Gorgolewski, K. J., "
    "Flandin, G., Phillips, C., Delorme, A., Oostenveld, R. (2019). "
    "EEG-BIDS, an extension to the brain imaging data structure "
    "for electroencephalography. Scientific Data, 6, 103."
    "https://doi.org/10.1038/s41597-019-0104-8",
    "ieeg": "Holdgraf, C., Appelhoff, S., Bickel, S., Bouchard, K., "
    "D'Ambrosio, S., David, O., … Hermes, D. (2019). iEEG-BIDS, "
    "extending the Brain Imaging Data Structure specification "
    "to human intracranial electrophysiology. Scientific Data, "
    "6, 102. https://doi.org/10.1038/s41597-019-0105-7",
    "nirs": "Luke, R., Oostenveld, R., Cockx, H., Niso, G., Shader, M., "
    "Orihuela-Espina, F., Innes-Brown, H., Tucker, S., Boas, D., Gau, R., "
    "Salo, T., Appelhoff, S., Markiewicz, C McAlpine, D., BIDS maintainers, "
    "Pollonini, L. (2023). fNIRS-BIDS, the Brain Imaging Data Structure "
    "Extended to Functional Near-Infrared Spectroscopy. PsyArXiv. "
    "https://doi.org/10.31219/osf.io/7nmcp",
}


# Mapping subject information between MNE-BIDS and MNE-Python.
HAND_BIDS_TO_MNE = {
    ("n/a",): 0,
    ("right", "r", "R", "RIGHT", "Right"): 1,
    ("left", "l", "L", "LEFT", "Left"): 2,
    ("ambidextrous", "a", "A", "AMBIDEXTROUS", "Ambidextrous"): 3,
}

HAND_MNE_TO_BIDS = {0: "n/a", 1: "R", 2: "L", 3: "A"}

SEX_BIDS_TO_MNE = {
    ("n/a", "other", "o", "O", "OTHER", "Other"): 0,
    ("male", "m", "M", "MALE", "Male"): 1,
    ("female", "f", "F", "FEMALE", "Female"): 2,
}

SEX_MNE_TO_BIDS = {0: "n/a", 1: "M", 2: "F"}


def _map_options(what, key, fro, to):
    if what == "sex":
        mapping_bids_mne = SEX_BIDS_TO_MNE
        mapping_mne_bids = SEX_MNE_TO_BIDS
    elif what == "hand":
        mapping_bids_mne = HAND_BIDS_TO_MNE
        mapping_mne_bids = HAND_MNE_TO_BIDS
    else:
        raise ValueError("Can only map `sex` and `hand`.")

    if fro == "bids" and to == "mne":
        # Many-to-one mapping
        mapped_option = None
        for bids_keys, mne_option in mapping_bids_mne.items():
            if key in bids_keys:
                mapped_option = mne_option
                break
    elif fro == "mne" and to == "bids":
        # One-to-one mapping
        mapped_option = mapping_mne_bids.get(key, None)
    else:
        raise RuntimeError(
            f"fro value {fro} and to value {to} are not "
            "accepted. Use 'mne', or 'bids'."
        )

    return mapped_option


# Which JSON data can safely be transferred from a non-anonymized to an
# anonymized dataset without accidentally exposing personal identifiable
# information
ANONYMIZED_JSON_KEY_WHITELIST = [
    # Common
    "Manufacturer",
    "ManufacturersModelName",
    "InstitutionName",
    "InstitutionalDepartmentName",
    "InstitutionAddress",
    "DeviceSerialNumber",
    # MRI
    # Many of these are not standardized, but produced by dcm2niix.
    "Modality",
    "MagneticFieldStrength",
    "ImagingFrequency",
    "StationName",
    "SeriesInstanceUID",
    "StudyInstanceUID",
    "StudyID",
    "BodyPartExamined",
    "PatientPosition",
    "ProcedureStepDescription",
    "SoftwareVersions",
    "MRAcquisitionType",
    "SeriesDescription",
    "ProtocolName",
    "ScanningSequence",
    "SequenceVariant",
    "ScanOptions",
    "SequenceName",
    "ImageType",
    "SeriesNumber",
    "AcquisitionNumber",
    "SliceThickness",
    "SAR",
    "EchoTime",
    "RepetitionTime",
    "InversionTime",
    "FlipAngle",
    "PartialFourier",
    "BaseResolution",
    "ShimSetting",
    "TxRefAmp",
    "PhaseResolution",
    "ReceiveCoilName",
    "ReceiveCoilActiveElements",
    "PulseSequenceDetails",
    "ConsistencyInfo",
    "PercentPhaseFOV",
    "PercentSampling",
    "PhaseEncodingSteps",
    "AcquisitionMatrixPE",
    "PixelBandwidth",
    "DwellTime",
    "ImageOrientationPatientDICOM",
    "InPlanePhaseEncodingDirectionDICOM",
    "ConversionSoftware",
    "ConversionSoftwareVersion",
    # Electrophys common
    "TaskName",
    "TaskDescription",
    "Instructions",
    "PowerLineFrequency",
    "SamplingFrequency",
    "SoftwareFilters",
    "RecordingType",
    "EEGChannelCount",
    "EOGChannelCount",
    "ECGChannelCount",
    "EMGChannelCount",
    "MiscChannelCount",
    "TriggerChannelCount",
    "RecordingDuration",
    # EEG
    "EEGReference",
    "EEGPlacementScheme",
    # MEG
    "DewarPosition",
    "DigitizedLandmarks",
    "DigitizedHeadPoints",
    "MEGChannelCount",
    "MEGREFChannelCount",
    "ContinuousHeadLocalization",
    "HeadCoilFrequency",
]



"""Command Line Interface for MNE-BIDS."""

# Authors: Teon Brooks <teon.brooks@gmail.com>
#          Stefan Appelhoff <stefan.appelhoff@mailbox.org>
#
# License: BSD-3-Clause
import glob
import os.path as op
import subprocess
import sys

import mne_bids

mne_bin_dir = op.abspath(op.dirname(mne_bids.__file__))
valid_commands = sorted(glob.glob(op.join(mne_bin_dir, "commands", "mne_bids_*.py")))
valid_commands = [c.split(op.sep)[-1][9:-3] for c in valid_commands]


def print_help():
    """Print the help."""
    print("Usage : mne_bids command options\n")
    print("Accepted commands :\n")
    for c in valid_commands:
        print("\t- %s" % c)
    print(
        "\nExample : mne_bids raw_to_bids --subject_id sub01 --task rest",
        "--raw_file data.edf --bids_root new_path",
    )
    sys.exit(0)


def main():
    """Run main command."""
    if len(sys.argv) == 1:
        print_help()
    elif "help" in sys.argv[1] or "-h" in sys.argv[1]:
        print_help()
    elif sys.argv[1] == "--version":
        print("MNE-BIDS %s" % mne_bids.__version__)
    elif sys.argv[1] not in valid_commands:
        print('Invalid command: "%s"\n' % sys.argv[1])
        print_help()
        sys.exit(0)
    else:
        cmd = sys.argv[1]
        cmd_path = op.join(mne_bin_dir, "commands", "mne_bids_%s.py" % cmd)
        sys.exit(subprocess.call([sys.executable, cmd_path] + sys.argv[2:]))



"""Write raw files to BIDS format.

example usage:  $ mne_bids report --bids_root bids_root_path

"""

# Authors: Alexandre Gramfort <alexandre.gramfort@inria.fr>
#
# License: BSD-3-Clause
import mne_bids
from mne_bids import make_report


def run():
    """Run the raw_to_bids command."""
    from mne.commands.utils import get_optparser

    parser = get_optparser(
        __file__,
        usage="usage: %prog options args",
        prog_prefix="mne_bids",
        version=mne_bids.__version__,
    )

    parser.add_option(
        "--bids_root", dest="bids_root", help="The path of the BIDS compatible folder."
    )

    opt, args = parser.parse_args()

    if len(args) > 0:
        parser.print_help()
        parser.error(f'Do not specify arguments without flags. Found: "{args}".\n')

    if not all([opt.bids_root]):
        parser.print_help()
        parser.error(
            "Arguments missing. You need to specify the --bids_root parameter."
        )

    report = make_report(opt.bids_root)
    print("-" * 36 + " REPORT " + "-" * 36)
    print(report)
    assert "  " not in report


if __name__ == "__main__":
    run()



"""Update BIDS directory structures and sidecar files meta data."""
# Authors: Adam Li <adam2392@gmail.com>
#          Austin Hurst <mynameisaustinhurst@gmail.com>
#          Stefan Appelhoff <stefan.appelhoff@mailbox.org>
#          mne-bids developers
#
# License: BSD-3-Clause

import json
from collections import OrderedDict

import numpy as np
from mne.channels import DigMontage, make_dig_montage
from mne.io import read_fiducials
from mne.io.constants import FIFF
from mne.utils import _check_on_missing, _on_missing, _validate_type, logger, verbose

from mne_bids import BIDSPath
from mne_bids.utils import _write_json


# TODO: add support for tsv files
@verbose
def update_sidecar_json(bids_path, entries, verbose=None):
    """Update sidecar files using a dictionary or JSON file.

    Will update metadata fields inside the path defined by
    ``bids_path.fpath`` according to the ``entries``. If a
    field does not exist in the corresponding sidecar file,
    then that field will be created according to the ``entries``.
    If a field does exist in the corresponding sidecar file,
    then that field will be updated according to the ``entries``.

    For example, if ``InstitutionName`` is not defined in
    the sidecar json file, then trying to update
    ``InstitutionName`` to ``Martinos Center`` will update
    the sidecar json file to have ``InstitutionName`` as
    ``Martinos Center``.

    Parameters
    ----------
    bids_path : BIDSPath
        The set of paths to update. The :class:`mne_bids.BIDSPath` instance
        passed here **must** have the ``.root`` attribute set. The
        ``.datatype`` attribute **may** be set. If ``.datatype`` is
        not set and only one data type (e.g., only EEG or MEG data)
        is present in the dataset, it will be
        selected automatically. This must uniquely identify
        an existing file path, else an error will be raised.
    entries : dict | str | pathlib.Path
        A dictionary, or JSON file that defines the
        sidecar fields and corresponding values to be updated to.
    %(verbose)s

    Notes
    -----
    This function can only update JSON files.

    Sidecar JSON files include files such as ``*_ieeg.json``,
    ``*_coordsystem.json``, ``*_scans.json``, etc.

    You should double check that your update dictionary is correct
    for the corresponding sidecar JSON file because it will perform
    a dictionary update of the sidecar fields according to
    the passed in dictionary overwriting any information that was
    previously there.

    Raises
    ------
    RuntimeError
        If the specified ``bids_path.fpath`` cannot be found
        in the dataset.

    RuntimeError
        If the ``bids_path.fpath`` does not have ``.json``
        extension.

    Examples
    --------
    Update a sidecar JSON file

    >>> from pathlib import Path
    >>> root = Path('./mne_bids/tests/data/tiny_bids').absolute()
    >>> bids_path = BIDSPath(subject='01', task='rest', session='eeg',
    ...                      suffix='eeg', extension='.json', datatype='eeg',
    ...                      root=root)
    >>> entries = {'PowerLineFrequency': 60}
    >>> update_sidecar_json(bids_path, entries, verbose=False)

    """
    # get all matching json files
    bids_path = bids_path.copy()
    if bids_path.extension != ".json":
        raise RuntimeError(
            'Only works for ".json" files. The '
            "BIDSPath object passed in has "
            f"{bids_path.extension} extension."
        )

    # get the file path
    fpath = bids_path.fpath
    if not fpath.exists():
        raise RuntimeError(f"Sidecar file does not " f"exist for {fpath}.")

    # sidecar update either from file, or as dictionary
    if isinstance(entries, dict):
        sidecar_tmp = entries
    else:
        with open(entries) as tmp_f:
            sidecar_tmp = json.load(tmp_f, object_pairs_hook=OrderedDict)

    logger.debug(sidecar_tmp)
    logger.debug(f"Updating {fpath}...")

    # load in sidecar filepath
    with open(fpath) as tmp_f:
        sidecar_json = json.load(tmp_f, object_pairs_hook=OrderedDict)

    # update sidecar JSON file with the fields passed in
    sidecar_json.update(**sidecar_tmp)

    # write back the sidecar JSON
    _write_json(fpath, sidecar_json, overwrite=True)


def _update_sidecar(sidecar_fname, key, val):
    """Update a sidecar JSON file with a given key/value pair.

    Parameters
    ----------
    sidecar_fname : str | os.PathLike
        Full name of the data file
    key : str
        The key in the sidecar JSON file. E.g. "PowerLineFrequency"
    val : str
        The corresponding value to change to in the sidecar JSON file.
    """
    with open(sidecar_fname, encoding="utf-8-sig") as fin:
        sidecar_json = json.load(fin)
    sidecar_json[key] = val
    _write_json(sidecar_fname, sidecar_json, overwrite=True)


@verbose
def update_anat_landmarks(
    bids_path,
    landmarks,
    *,
    fs_subject=None,
    fs_subjects_dir=None,
    kind=None,
    on_missing="raise",
    verbose=None,
):
    """Update the anatomical landmark coordinates of an MRI scan.

    This will change the ``AnatomicalLandmarkCoordinates`` entry in the
    respective JSON sidecar file, or create it if it doesn't exist.

    Parameters
    ----------
    bids_path : BIDSPath
        Path of the MR image.
    landmarks : mne.channels.DigMontage | path-like
        An :class:`mne.channels.DigMontage` instance with coordinates for the
        nasion and left and right pre-auricular points in MRI voxel
        coordinates. Alternatively, the path to a ``*-fiducials.fif`` file as
        produced by the MNE-Python coregistration GUI or via
        :func:`mne.io.write_fiducials`.

        .. note:: :func:`mne_bids.get_anat_landmarks` provides a convenient and
                  reliable way to generate the landmark coordinates in the
                  required coordinate system.

        .. note:: If ``path-like``, ``fs_subject`` and ``fs_subjects_dir``
                  must be provided as well.

        .. versionchanged:: 0.10
           Added support for ``path-like`` input.
    fs_subject : str | None
        The subject identifier used for FreeSurfer. Must be provided if
        ``landmarks`` is ``path-like``; otherwise, it will be ignored.
    fs_subjects_dir : path-like | None
        The FreeSurfer subjects directory. If ``None``, defaults to the
        ``SUBJECTS_DIR`` environment variable. Must be provided if
        ``landmarks`` is ``path-like``; otherwise, it will be ignored.
    kind : str | None
        The suffix of the anatomical landmark names in the JSON sidecar.
        A suffix might be present e.g. to distinguish landmarks between
        sessions. If provided, should not include a leading underscore ``_``.
        For example, if the landmark names in the JSON sidecar file are
        ``LPA_ses-1``, ``RPA_ses-1``, ``NAS_ses-1``, you should pass
        ``'ses-1'`` here.
        If ``None``, no suffix is appended, the landmarks named
        ``Nasion`` (or ``NAS``), ``LPA``, and ``RPA`` will be used.

        .. versionadded:: 0.10
    on_missing : 'ignore' | 'warn' | 'raise'
        How to behave if the specified landmarks cannot be found in the MRI
        JSON sidecar file.

        .. versionadded:: 0.10
    %(verbose)s

    Notes
    -----
    .. versionadded:: 0.8
    """
    _validate_type(item=bids_path, types=BIDSPath, item_name="bids_path")
    _validate_type(
        item=landmarks, types=(DigMontage, "path-like"), item_name="landmarks"
    )
    _check_on_missing(on_missing)

    # Do some path verifications and fill in some gaps the users might have
    # left (datatype and extension)
    # XXX We could be more stringent (and less user-friendly) and insist on a
    # XXX full specification of all parts of the BIDSPath, thoughts?
    bids_path_mri = bids_path.copy()
    if bids_path_mri.datatype is None:
        bids_path_mri.datatype = "anat"

    if bids_path_mri.datatype != "anat":
        raise ValueError(
            f'Can only operate on "anat" MRI data, but the provided bids_path '
            f"points to: {bids_path_mri.datatype}"
        )

    if bids_path_mri.suffix is None:
        raise ValueError(
            'Please specify the "suffix" entity of the provided ' "bids_path."
        )
    elif bids_path_mri.suffix not in ("T1w", "FLASH"):
        raise ValueError(
            f'Can only operate on "T1w" and "FLASH" images, but the bids_path '
            f"suffix indicates: {bids_path_mri.suffix}"
        )

    valid_extensions = (".nii", ".nii.gz")
    tried_paths = []
    file_exists = False
    if bids_path_mri.extension is None:
        # No extension was provided, start searching …
        for extension in valid_extensions:
            bids_path_mri.extension = extension
            tried_paths.append(bids_path_mri.fpath)

            if bids_path_mri.fpath.exists():
                file_exists = True
                break
    else:
        # An extension was provided
        tried_paths.append(bids_path_mri.fpath)
        if bids_path_mri.fpath.exists():
            file_exists = True

    if not file_exists:
        raise ValueError(
            f"Could not find an MRI scan. Please check the provided "
            f"bids_path. Tried the following filenames: "
            f'{", ".join([p.name for p in tried_paths])}'
        )

    if not isinstance(landmarks, DigMontage):  # it's pathlike
        if fs_subject is None:
            raise ValueError(
                'You must provide the "fs_subject" parameter when passing the '
                "path to fiducials"
            )
        landmarks = _get_landmarks_from_fiducials_file(
            bids_path=bids_path,
            fname=landmarks,
            fs_subject=fs_subject,
            fs_subjects_dir=fs_subjects_dir,
        )

    positions = landmarks.get_positions()
    coord_frame = positions["coord_frame"]
    if coord_frame != "mri_voxel":
        raise ValueError(
            f"The landmarks must be specified in MRI voxel coordinates, but "
            f'provided DigMontage is in "{coord_frame}"'
        )

    # Extract the cardinal points
    name_to_coords_map = {
        "LPA": positions["lpa"],
        "NAS": positions["nasion"],
        "RPA": positions["rpa"],
    }

    # Check if coordinates for any cardinal point are missing, and convert to
    # a list so we can easily store the data in JSON format
    missing_points = []
    for name, coords in name_to_coords_map.items():
        if coords is None:
            missing_points.append(name)
        else:
            # Funnily, np.float64 is JSON-serializabe, while np.float32 is not!
            # Thus, cast to float64 to avoid issues (which e.g. may arise when
            # fiducials were read from disk!)
            name_to_coords_map[name] = list(coords.astype("float64"))

    if missing_points:
        raise ValueError(
            f"The provided DigMontage did not contain all required cardinal "
            f"points (nasion and left and right pre-auricular points). The "
            f"following points are missing: "
            f'{", ".join(missing_points)}'
        )

    bids_path_json = bids_path.copy().update(extension=".json")
    if not bids_path_json.fpath.exists():  # Must exist before we can update it
        _write_json(bids_path_json.fpath, dict())

    mri_json = json.loads(bids_path_json.fpath.read_text(encoding="utf-8"))
    if "AnatomicalLandmarkCoordinates" not in mri_json:
        _on_missing(
            on_missing=on_missing,
            msg=f"No AnatomicalLandmarkCoordinates section found in "
            f"{bids_path_json.fpath.name}",
            error_klass=KeyError,
        )
        mri_json["AnatomicalLandmarkCoordinates"] = dict()

    for name, coords in name_to_coords_map.items():
        if kind is not None:
            name = f"{name}_{kind}"

        if name not in mri_json["AnatomicalLandmarkCoordinates"]:
            _on_missing(
                on_missing=on_missing,
                msg=f"Anatomical landmark not found in "
                f"{bids_path_json.fpath.name}: {name}",
                error_klass=KeyError,
            )

        mri_json["AnatomicalLandmarkCoordinates"][name] = coords

    update_sidecar_json(bids_path=bids_path_json, entries=mri_json)


def _get_landmarks_from_fiducials_file(
    *, bids_path, fname, fs_subject, fs_subjects_dir
):
    """Get anatomical landmarks from fiducials file, in MRI voxel space."""
    # avoid dicrular imports
    from mne_bids.write import (
        _get_fid_coords,
        _get_t1w_mgh,
        _mri_landmarks_to_mri_voxels,
    )

    digpoints, coord_frame = read_fiducials(fname)

    # All of this should be guaranteed, but better be safe than sorry!
    assert coord_frame == FIFF.FIFFV_COORD_MRI
    assert digpoints[0]["ident"] == FIFF.FIFFV_POINT_LPA
    assert digpoints[1]["ident"] == FIFF.FIFFV_POINT_NASION
    assert digpoints[2]["ident"] == FIFF.FIFFV_POINT_RPA

    montage_loaded = make_dig_montage(
        lpa=digpoints[0]["r"],
        nasion=digpoints[1]["r"],
        rpa=digpoints[2]["r"],
        coord_frame="mri",
    )
    landmark_coords_mri, _ = _get_fid_coords(dig_points=montage_loaded.dig)
    landmark_coords_mri = np.asarray(
        (
            landmark_coords_mri["lpa"],
            landmark_coords_mri["nasion"],
            landmark_coords_mri["rpa"],
        )
    )

    t1w_mgh = _get_t1w_mgh(fs_subject, fs_subjects_dir)
    landmark_coords_voxels = _mri_landmarks_to_mri_voxels(
        mri_landmarks=landmark_coords_mri * 1000,
        t1_mgh=t1w_mgh,  # in mm
    )
    montage_voxels = make_dig_montage(
        lpa=landmark_coords_voxels[0],
        nasion=landmark_coords_voxels[1],
        rpa=landmark_coords_voxels[2],
        coord_frame="mri_voxel",
    )

    return montage_voxels



"""Some functions to extract stats from a BIDS dataset."""

# Authors: Alex Gramfort <alexandre.gramfort@inria.fr>
#
# License: BSD-3-Clause

from mne_bids import BIDSPath, get_datatypes
from mne_bids.config import EPHY_ALLOWED_DATATYPES


def count_events(root_or_path, datatype="auto"):
    """Count events present in dataset.

    Parameters
    ----------
    root_or_path : path-like | mne_bids.BIDSPath
        If str or Path it is the root folder of the BIDS dataset.
        If a BIDSPath is passed it allows to limit the count
        to a subject, a session or a run by only considering
        the event files that match this BIDSPath.
    datatype : str
        Type of the data recording. Can be ``meg``, ``eeg``,
        ``ieeg`` or ``auto``. If ``auto`` and a :class:`mne_bids.BIDSPath`
        isinstance is passed as ``root_or_path`` which has a ``datatype``
        attribute set, then this data type will be used. Otherwise, only
        one data type should be present in the dataset to avoid any
        ambiguity.

    Returns
    -------
    counts : pandas.DataFrame
        The pandas dataframe containing all the counts of trial_type
        in all matching events.tsv files.

    Notes
    -----
    .. versionchanged:: 0.15
       Table values were changed from floats (with NaN for missing values)
       to Pandas nullable integer arrays.
    """
    import pandas as pd

    if not isinstance(root_or_path, BIDSPath):
        bids_path = BIDSPath(root=root_or_path)
    else:
        bids_path = root_or_path.copy()

    bids_path.update(suffix="events", extension=".tsv")

    datatypes = get_datatypes(bids_path.root)
    this_datatypes = list(set(datatypes).intersection(EPHY_ALLOWED_DATATYPES))

    if (datatype == "auto") and (bids_path.datatype is not None):
        datatype = bids_path.datatype

    if datatype == "auto":
        if len(this_datatypes) > 1:
            raise ValueError(
                f"Multiple datatypes present ({this_datatypes})."
                f" You need to specity datatype got: {datatype})"
            )
        elif len(this_datatypes) == 0:
            raise ValueError("No valid datatype present.")

        datatype = this_datatypes[0]

    if datatype not in EPHY_ALLOWED_DATATYPES:
        raise ValueError(
            f"datatype ({datatype}) is not supported. "
            f"It must be one of: {EPHY_ALLOWED_DATATYPES})"
        )

    bids_path.update(datatype=datatype)

    tasks = sorted(set([bp.task for bp in bids_path.match()]))

    all_counts = []

    for task in tasks:
        bids_path.update(task=task)

        all_df = []
        for bp in bids_path.match():
            df = pd.read_csv(str(bp), delimiter="\t")
            df["subject"] = bp.subject
            if bp.session is not None:
                df["session"] = bp.session
            if bp.run is not None:
                df["run"] = bp.run
            all_df.append(df)

        if not all_df:
            continue

        df = pd.concat(all_df)
        groups = ["subject"]
        if bp.session is not None:
            groups.append("session")
        if bp.run is not None:
            groups.append("run")

        if "stim_type" in df.columns:
            # Deal with some old files that use stim_type rather than
            # trial_type
            df = df.rename(columns={"stim_type": "trial_type"})

        # There are datasets out there without a `trial_type` or `stim_type`
        # column.
        if "trial_type" in df.columns:
            groups.append("trial_type")

        counts = df.groupby(groups).size()
        counts = counts.unstack(fill_value=-1)
        counts.replace(-1, pd.NA, inplace=True)

        if "BAD_ACQ_SKIP" in counts.columns:
            counts = counts.drop("BAD_ACQ_SKIP", axis=1)

        counts.columns = pd.MultiIndex.from_arrays(
            [[task] * counts.shape[1], counts.columns]
        )

        all_counts.append(counts)

    if not all_counts:
        raise ValueError("No events files found.")

    counts = pd.concat(all_counts, axis=1)

    return counts
